#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright © 2021 Adam Twardoch, MIT license
# Adapted from : https://github.com/twardoch/gimeltra

# Modified by Vinodh Rajan

import os
import regex as re
from pathlib import Path
import logging
import json
import langcodes
from fontTools import unicodedata as ucd
from collections import Counter

cwd = Path(Path(__file__).parent)

class Transliterator(object):
    def __init__(self):
        with open(Path(cwd, "json/gimeltra_data.json"), "r", encoding="utf-8") as f:
            data = json.load(f)
        self.db = data['ssub']
        self.db_ccmp = data['ccmp']
        self.db_simplify = data['simp']['general']
        self.db_fina = data['fina']
        self.db_liga = data['liga']
        self.vocalized = ['Hebr', 'Hebr-Ar', 'Syrj', 'Syrn', 'Arab-Fa', 'Arab-Pa', 'Latn', 'Latn-No', 'Type', 'Arab', 'Arab-Ur', 'Thaa']

    def auto_script(self, text):
        sc_count = Counter([ucd.script(c) for c in text])
        sc = sc_count.most_common(1)[0][0]
        if not sc:
            sc = 'Zyyy'
        return sc

    def _tr(self, text, sc, to_sc):
        t = text
        if sc != 'Latn':
            t = self._preprocess(t, sc, to_sc)
        t = self._convert(t, sc, to_sc)
        t = self._postprocess(t, to_sc)
        return t

    def _preprocess(self, text, sc, to_sc):
        t = text
        for rule_i, rule_o in self.db_ccmp.get(sc, {}).items():
            t = t.replace(rule_i, rule_o)

        if sc not in self.vocalized or to_sc not in self.vocalized:
            t = ucd.normalize('NFD', t)
            t = re.sub(r"(?![\u05C4\u0308])\p{M}","", t)

        logging.debug(f"Pre: {list(t)}")
        return t

    def _postprocess(self, text, sc):
        t = text
        for rule_i, rule_o in self.db_fina.get(sc, {}).items():
            #t = re.subf(fr"(\p{{L}})({rule_i})([^\p{{L}}])", f"{{1}}{rule_o}{{3}}", t)
            #t = re.subf(fr"(\p{{L}})({rule_i})$", f"{{1}}{rule_o}", t)

            t = re.subf(fr"(\p{{L}})(\p{{M}}*?)({rule_i})(\p{{M}}*?)([^\p{{L}}])", f"{{1}}{{2}}{rule_o}{{4}}{{5}}", t)
            t = re.subf(fr"(\p{{L}})(\p{{M}}*?)({rule_i})(\p{{M}}*?)$", f"{{1}}{{2}}{rule_o}{{4}}", t)
            t = re.subf(fr"({rule_o})(\p{{M}}+)(\p{{L}})", f"{rule_i}{{2}}{{3}}", t)


        for rule_i, rule_o in self.db_liga.get(sc, {}).items():
            t = t.replace(rule_i, rule_o)
        logging.debug(f"Post: {list(t)}")

        return t

    def _to_latin(self, text, sc, to_sc):
        chars = list(self.db[sc]["Latn"])
        chars.sort(key=len, reverse=True)
        #print(chars)
        for char in chars:
            #print(char + ' : ' + self.db[sc]["Latn"][char])
            text = text.replace(char, self.db[sc]["Latn"][char])
            #print(text)

        # print(text)
        return text

    def _from_latin(self, text, sc, to_sc):
        chars = list(self.db["Latn"][to_sc])
        chars.sort(key=len, reverse=True)

        #if sc != 'Latn':
            # chars_missing = set(list(self.db[sc]["Latn"].values())) - set(chars)

        #if sc == 'Latn':
            #chars_missing = set(self.db_simplify) -  set(chars) #set(list(self.db[to_sc]["Latn"].values()))
            #print(chars)

        chars_missing = [x for x in self.db_simplify if x not in chars]
        #chars_missing = list(set(self.db_simplify) -  set(chars))

        for char in chars_missing:
            if char in self.db_simplify:
                #print(char, self.db_simplify[char])
                ## if Virama is misssing and characts has virama remove virama
                if '\u033D' in chars_missing and '\u033D' in self.db_simplify[char]:
                    #print('here removing vriama')
                    #print(char, self.db_simplify[char])
                    text = text.replace(char, self.db_simplify[char])
                    text = text.replace('\u033D', self.db_simplify['\u033D'])
                else:
                    text = text.replace(char, self.db_simplify[char])


        for char in chars:
            #print(text)
            #print(char + " " + self.db["Latn"][to_sc][char])
            text = text.replace(char, self.db["Latn"][to_sc][char])
            #print(text)

        return text

    def _convert(self, text, sc, to_sc):
        if to_sc == 'Latn':
            return self._to_latin(text, sc, to_sc)
        elif sc == 'Latn':
            return self._from_latin(text, sc, to_sc)
        else:
            txt_latn = self._to_latin(text, sc, to_sc)
            return self._from_latin(txt_latn, sc, to_sc)

    def tr(self, text, sc=None, to_sc='Latn'):
        if not sc:
            sc = self.auto_script(text)
        logging.debug({
            'script': sc, 'to_script': to_sc,
        })
        logging.debug(f"Text: {list(text)}")
        if sc != to_sc:
            res = self._tr(text, sc, to_sc)
        else:
            res = text
        return res

def tr(text, sc=None, to_sc='Latn'):
    tr = Transliterator()
    # print(sc +' ' + to_sc)
    if sc != to_sc:
        return tr.tr(text, sc, to_sc)
    else:
        return text




# -*- coding: utf-8 -*-

# Script Mapping for ISO


VowelMap =  [chr(0xE000+i) for i in range(0,14)]

SouthVowelMap = [chr(0xE100+i) for i in range(0,2)]

ModernVowelMap = [chr(0xE200+i) for i in range(0,2)]

SinhalaVowelMap = [chr(0xE300+i) for i in range(0,1)]

VowelSignMap =  VowelMap[1:]

SouthVowelSignMap = SouthVowelMap[:]

ModernVowelSignMap = ModernVowelMap[:]

SinhalaVowelSignMap = SinhalaVowelMap[:]

AyogavahaMap = [chr(0xE400+i) for i in range(0,3)]

ViramaMap =  [
             '\u00D7'
             ]

ConsonantMap =  [chr(0xE500+i) for i in range(0,33)]

SouthConsonantMap = [chr(0xE600+i) for i in range(0,4)]

NuktaConsonantMap =  [chr(0xE700+i) for i in range(0,8)]

NuktaMap = [chr(0xE800+i) for i in range(0,1)]

OmMap = [chr(0xE900+i) for i in range(0,1)]

SignMap = [chr(0xEA00+i) for i in range(0,3)]

SinhalaConsonantMap =[chr(0xEB00+i) for i in range(0,5)]

Aytham = [chr(0xEC00+i) for i in range(0,1)]

NumeralMap = [chr(0xED00+i) for i in range(0,10)]

from ... import GeneralMap as GM

import os
GM.add_additional_chars(dict([(charlist, globals()[charlist]) for charlist in GM.CharmapLists]),
                        os.path.splitext(__file__)[0].split(os.path.sep)[-1])



# -*- coding: utf-8 -*-

from ast import Str
from asyncio import constants
from . import GeneralMap as GM
import re
import string
import unicodedata
from . import PostProcess
from . import ConvertFix as CF
from aksharamukha.ScriptMap.EastIndic import PhagsPa, Burmese, Khmer
from aksharamukha.ScriptMap.MainIndic import Tamil, Malayalam, Limbu, Chakma
### Use escape char in all functions

def ItransLL(Strng):
    return Strng

def RomanLoCChandrabindu(Strng):
    Strng = re.sub('n̐', 'm̐', Strng)

    return Strng

def BalineseRomanLoCFix(Strng):
    Strng = Strng.replace('‘', 'ng̈')

    return Strng

def JavaneseRomanLoCFix(Strng):
    Strng = Strng.replace('‘', 'ng̈')

    return Strng

def HindiMarathiRomanLoCFix(Strng):
    Strng = Strng.replace('sh', 'ṣ')
    Strng = Strng.replace('ḷ', 'ḻ')
    Strng = Strng.replace('l̤', 'l̳')

    return Strng

def RomanLoCLaUnderscoreDoubleDot(Strng):
    Strng = Strng.replace('ḻ', 'l̤')

# reverse these in preprocess
def RomanLoCVaWa(Strng):
    Strng = Strng.replace('w', 'v')
    return Strng

def RomanLoCSasha(Strng):
    Strng = Strng.replace('sh', 'ṣ')

    return Strng

def RomanLoCSLaDotLaUnderscore(Strng):
    Strng = Strng.replace('ḻ', 'l̤')
    Strng = Strng.replace( 'l̳', 'ḻ',)

    return Strng

def RomanLoCLaUnderscoreDoubleDot(Strng):
    Strng = Strng.replace('l̤', 'ḻ')

    return Strng

def MalayalamRomanLoCFix(Strng):
    Strng = Strng.replace('ṯṯ', 'ṟṟ')
    Strng = Strng.replace('ȧ', 'ŭ')
    return Strng

def DevanagariRomanLoCFix(Strng):
    Strng = Strng.replace('g̳h̳', 'gḧ').replace('t̤', 'ṭ̈').replace('s̤', 's̈')\
        .replace('h̤', 'ḧ')

    return Strng

def IASTLDotRetroflex(Strng):
    Strng = Strng.replace('ḷ', 'l̤')

    return Strng

def ArchaicJnaSimplifyRomanLOC(Strng):
    Strng = Strng.replace('ꦘ', 'ꦚ')
    return Strng

def KhandaTaRomanLoC(Strng):
    Strng = Strng.replace('ৎ', 'ṯ')

    return Strng

def TibetanLoCRomanLoCFix(Strng):
    Strng = re.sub('t(?![ʹhs])', 'tʹ', Strng)
    Strng = re.sub('n(?![ʹy])', 'nʹ', Strng)

    return Strng

def BalineseJavaneseMoveRepha(Strng, tgt, reph):
    repha = '(' + reph + ')'

    cons = '(' + '|'.join(GM.CrunchSymbols(GM.Consonants, tgt)) + ')'
    vows = '(' + '|'.join(GM.CrunchSymbols(GM.VowelSignsNV, tgt)) + ')'
    vir = GM.CrunchSymbols(GM.virama, tgt)[0]

    candAnu = '[' + ''.join(GM.CrunchSymbols(GM.CombiningSigns, tgt)[:2]) + ']'
    Strng = re.sub('(' + candAnu + ')' + repha, r'\2\1', Strng)

    Strng = re.sub('('+cons+')'+'('+ vir + cons +')*'+'(' + vows + ')?'+ repha, r'\7\1\3\5', Strng)

    return Strng

def DivesAkuruAlternateIndVowels(Strng):
    # use alt /y/
    #Strng = Strng.replace("\U00011925", "\U00011926")

    # replace ind. vow with /y/

    vow = "𑤁 𑤂 𑤃 𑤄 𑤅 𑤆 𑤆𑤵 𑤉 𑤀".split(" ")
    vowy = "𑤥𑤰 𑤥𑤱 𑤥𑤲 𑤥𑤳 𑤥𑤴 𑤥𑤵 𑤥𑤷 𑤥𑤸 𑤥".split(" ")

    for v, vy in zip(vow, vowy):
        Strng = Strng.replace(vy, v)

    return Strng

def KawiMoveRepha(Strng):
    tgt = 'Kawi'
    repha = '\U00011F02'

    cons = '(' + '|'.join(GM.CrunchSymbols(GM.Consonants + GM.Vowels, tgt)) + ')'
    vows = '(' + '|'.join(GM.CrunchSymbols(GM.VowelSignsNV, tgt)) + ')'
    vir = '\U00011F42'

    Strng = re.sub(repha + '(('+cons+')' + '('+ vir + cons +')*'+'(' + vows + ')?)', r'\1' + repha, Strng)

    return Strng

def JavaneseMoveRepha(Strng):
    return BalineseJavaneseMoveRepha(Strng, 'Javanese', 'ꦂ')

def BalineseMoveRepha(Strng):
    return BalineseJavaneseMoveRepha(Strng, 'Balinese', 'ᬃ')

def OriyaSubojinedVa(Strng):
    Strng = re.sub('(??<![ମବ])(୍ବ)', '୍ୱ', Strng)

    return Strng

def BengaliSubojinedVa(Strng):
    Strng = re.sub('(?<![মব])(্ব)', '্ভ়', Strng)

    return Strng

def BengaliTargetVa(Strng):
    Strng = Strng.replace('ব', 'ভ়')

    return Strng

def OriyaTargetVa(Strng):
    Strng = Strng.replace('ବ', 'ୱ')

    return Strng

def RetainDevangariDanda(Strng):
    Strng = Strng.replace('।', '│').replace('॥', '┃')
    return Strng

def RetainPipeDanda(Strng, src):
    Strng = Strng.replace("।", "|").replace("॥", "||")
    Strng = Strng.replace(".", "●")
    Strng = Strng.replace("||", "┃").replace("|", "│")

    return Strng

#ThamLoC
def ThamLoCRomanLoCTarget(Strng):
    ListC ='|'.join(GM.CrunchSymbols(GM.Consonants,'ThamLoC'))

    #Fix o
    Strng = Strng.replace('ᩮᩣ', 'ᩰ')

    #eo
    Strng = re.sub('ᩴ᩠ᨿ', '\u1A74\u1A7Fᨿ', Strng)

    #aiy
    Strng = re.sub('ᩱ᩠ᨿ', 'ᩱ\u1A7Fᨿ', Strng)

    #koi
    Strng = re.sub('᩠ᩅᩭ', '\u1A7Fᩅᩭ', Strng)

    #au
    Strng = re.sub('\u1A60\u1A45\u1A6B', '\u1A7F\u1A45\u1A6B', Strng)

    #close e/au
    Strng = re.sub('(᩠)(ᨿ|ᩅ)(?=(' + ListC + '))', '\u1A7F'+r'\2', Strng)


    return Strng

def ThamLoCRomanLoCSource(Strng):
    ListC ='|'.join(GM.CrunchSymbols(GM.Consonants,'ThamLoCRomanLoC'))
    ListV ='|'.join(GM.CrunchSymbols(GM.Vowels,'ThamLoCRomanLoC'))

    #a
    Strng = re.sub('(a|ǫḥ|ǫ|œ|ò)(?=(' + ListC + '))', r'\1''\u02BD', Strng)

    return Strng

def KhmerWordSplit(Strng):
    from khmernltk import word_tokenize
    sents = Strng.split('\n')
    sent_token = []
    for sent in sents:
        sent_token.append(' '.join(word_tokenize(sent)))
    Strng = '\n'.join(sent_token)
    Strng = Strng.replace('  ', ' ')

    return Strng

def KhmerLoCRomanLoCTarget(Strng):
    ListC ='|'.join(GM.CrunchSymbols(GM.Consonants,'Khmer'))
    ListV ='|'.join(GM.CrunchSymbols(GM.VowelSigns,'Khmer'))
    ListA ='|'.join(GM.CrunchSymbols(GM.CombiningSigns,'Khmer'))

    vir = Khmer.ViramaMap[0]

    #Mark im
    Strng = Strng.replace('ឹ', 'ិំ\u02BD')

    #move bantoc around
    Strng = re.sub('(.)(.(\u17D2.)*)(\u17CB)', r'\1\4\2', Strng)

    #mark consonant modifiers with subbase form
    Strng = re.sub('([ងញមបយរវ])(ុ)([ិឹីឺ])', r'\1''៉'r'\3', Strng)
    Strng = re.sub('([សហអ])(ុ)([ិឹីឺ])', r'\1''៊'r'\3', Strng)

    #move tonemarks around
    Strng = re.sub('('+ListC+')'+'([៎៏])''('+ListV+')''('+ListA+')?',r'\1\3\4\2',Strng)
    Strng = re.sub('('+ListC+')'+'('+ListV+')''([៎៏៊៉])''('+ListA+')?',r'\1\2\4\3',Strng)

    #fix oya to oy
    Strng = Strng.replace('ឲ្យ', 'ឱ្យ').replace('ឱ្យ', 'ឱ្យ' + vir)

    #ignore the following
    Strng = re.sub('[៙៚]', '', Strng)

    return Strng

def KhmerLoCRomanLoCSource(Strng):
    listdenormalize = ['å', 'ẙ', 'à', 'á', 'é', 'í', 'ó', 'ú', 'à', 'è', 'ì', 'ò', 'ù', 'ă']

    for let in listdenormalize:
        Strng = Strng.replace(let, unicodedata.normalize('NFD', let))

    Strng = re.sub('‛(?!ʹ)', '‛ʹ', Strng)
    Strng = re.sub('oaḥ', 'oḥ', Strng)

    return Strng

# shan
def ShanLoCRomanLoCTarget(Strng):
    #preserve ႂ်
    Strng = Strng.replace('ႂ်', '\u036E')

    # Shan normalize /aa/
    Strng = Strng.replace('ၢ', 'ႃ')
    Strng = Strng.replace('ါ', 'ႃ')

    Strng = Strng.replace('ႃႆ', 'ၢႆ').replace('ႆၢ', 'ၢႆ').replace('ႆႃ', 'ၢႆ')

    # asat + virma to just virama
    Strng = Strng.replace('\u103A\u1039', '\u1039')

    ## sort subjoined consonants
    yrv = Burmese.ConsonantMap[25:27] + Burmese.ConsonantMap[28:29]
    yrvsub = ['\u103B','\u103C','\u1082']
    vir = Burmese.ViramaMap[0]

    for x,y in zip(yrv,yrvsub):
        # Undo Replace subjoining forms: exp-virama + y/r/v/h <- subjoining y/r/v/h
        Strng = Strng.replace(y, vir + vir + x)

    # mark pure viramas
    aThat = r'်'
    Strng = re.sub('(?<!ႃ)'+aThat, aThat + 'ʻ', Strng)

    # mark a as glottalstop
    Strng = Strng.replace('ဢ','’ဢ')

    return Strng

def ShanLoCRomanLoCSource(Strng):

    return Strng

def BurmeseRomanLoCSource(Strng):
    # adhoc chars
    chars_misc = {
        "e*": "၏",
        'n*': "၌",
        'r*': '၍',
        'l*': '၎'
    }

    for lat, bur in chars_misc.items():
        Strng = Strng.replace(lat, bur)

    Strng = Strng.replace('ṁ', 'ṃ')

    # reverse danda to comma and double danda to full stop
    Strng = Strng.replace(',', '၊').replace('.', '။')

    # restore visarga and fix incorrect diacritic character
    Strng = Strng.replace('˝', 'ʺ').replace('ʺ', 'ḥ')

    # restore visarga and fix incorrect diacritic character
    Strng = Strng.replace('´','ʹ').replace('ʹ', '˳')

    # replace modifier letter with quotation mark
    vowelSigns = '|'.join(GM.CrunchSymbols(GM.VowelSignsNV, 'IAST'))
    Strng = re.sub('(ʼ)(a|' + vowelSigns + ')', '’' + r'\2', Strng)

    # left quotation mark -> modifier letter comma
    consonants = '|'.join(GM.CrunchSymbols(GM.Consonants, 'IAST'))
    Strng = re.sub('(' + consonants + ')(‘)', r'\1' + 'ʻ', Strng)

    Strng = Strng.replace('o‘', 'oʻ')

    # reverse o' to au
    Strng = Strng.replace('oʻ', 'au')

    return Strng

def paliTham(Strng):

    return Strng

def segmentThamSyllabes(Strng):
    # segment text into syllables

    # https://github.com/ye-kyaw-thu/myWord/blob/main/syl_segment.py
    myConsonant = r"ᨠ-ᩌ"
    otherChar = r"ᩍ-ᩔ!-/:-@[-`{-~\s"
    ssSymbol = r'᩠'
    aThat = r'᩺'

    BreakPattern = re.compile(r"((?<!" + ssSymbol + r")["+ myConsonant + r"](?![" + aThat + ssSymbol + r"])" + r"|["  + otherChar + r"])", re.UNICODE)
    #Strng = Strng.replace("့်", "့်")
    Strng = BreakPattern.sub(' ' + r"\1", Strng)

    return Strng

def segmentBurmeseSyllables(Strng):
    # segment text into syllables

    # https://github.com/ye-kyaw-thu/myWord/blob/main/syl_segment.py
    myConsonant = r"က-အ"
    otherChar = r"ဣဤဥဦဧဩဪဿ၌၍၏၀-၉၊။!-/:-@[-`{-~\s"
    ssSymbol = r'္'
    aThat = r'်'

    BreakPattern = re.compile(r"((?<!" + ssSymbol + r")["+ myConsonant + r"](?![" + aThat + ssSymbol + r"])" + r"|["  + otherChar + r"])", re.UNICODE)
    Strng = Strng.replace("့်", "့်")
    Strng = BreakPattern.sub(' ' + r"\1", Strng)

    return Strng

def segmentShanSyllables(Strng):
    # segment text into syllables

    # https://github.com/ye-kyaw-thu/myWord/blob/main/syl_segment.py
    myConsonant = r"ၵၶၷꧠငၸꧡꩡꧢၺꩦꩧꩨꩩꧣတထၻꩪၼပၽၿꧤမယရလဝသႁꩮၹၾႀဢ"
    otherChar = r"႞႟႐-႙၊။!-/:-@[-`{-~\s"
    ssSymbol = r'္'
    aThat = r'်'

    BreakPattern = re.compile(r"((?<!" + ssSymbol + r")["+ myConsonant + r"](?![" + aThat + ssSymbol + r"])" + r"|["  + otherChar + r"])", re.UNICODE)
    Strng = Strng.replace("့်", "့်")
    Strng = BreakPattern.sub(' ' + r"\1", Strng)

    return Strng

def BurmeseRomanLoCTarget(Strng):
    # asat + virma to just virama
    Strng = Strng.replace('\u103A\u1039', '\u1039')

    # swap iu -> ui
    Strng = Strng.replace('\u102D\u102F', '\u102F\u102D')

    # mark Independent au -> o'
    Strng = Strng.replace('ဪ','ဩʻ')

    ## sort subjoined consonants
    yrvh = Burmese.ConsonantMap[25:27] + Burmese.ConsonantMap[28:29] + Burmese.ConsonantMap[32:33]
    yrvhsub = ['\u103B','\u103C','\u103D','\u103E']
    vir = Burmese.ViramaMap[0]

    for x,y in zip(yrvh,yrvhsub):
        # Undo Replace subjoining forms: exp-virama + y/r/v/h <- subjoining y/r/v/h
        Strng = Strng.replace(y, vir + vir + x)

    # swap tone + virama to proper order
    Strng = Strng.replace("့်", "့်")

    # mark pure viramas
    aThat = r'်'
    Strng = Strng.replace(aThat, aThat + 'ʻ')

    # mark a as glottalstop
    Strng = Strng.replace('အ','’အ')

    # replace vowel + diac with vowels
    vowDep = 'အော် အော အိ အီ အု အူ အေ'.split(' ')
    vowIndep = 'ဪ ဩ ဣ ဤ ဥ ဦ ဧ'.split(' ')

    for x, y in zip(vowDep, vowIndep):
        Strng = Strng.replace(x, y)

    # danda to comma and double danda to full stop
    Strng = Strng.replace('၊', ',').replace('။', '.')

    #print(Strng)

    return Strng

def insertViramaSyriac(Strng):
    Strng += "\uF001"
    return Strng

def BengaliSwitchYaYYa(Strng):
    Strng = re.sub('(?<!\u09CD)য', '@#$', Strng)
    Strng = re.sub('য়', 'য', Strng)
    Strng = Strng.replace('@#$', 'য়')

    return Strng


def removeFinalSchwaArab(Strng):
    #print('here', Strng)
    diacrtics = ["\u0652", "\u064E", "\u0650", "\u064F"]
    Strng = re.sub('([\u0628-\u0647])(?![\u0652\u064E\u0650\u064F\u0651\u064B\u064C\u064D\u0649])(?=(\W|$))', r'\1' + '\u0652', Strng)
    Strng = re.sub('([\u0628-\u0647]\u0651)(?![\u0652\u064E\u0650\u064F\u064B\u064C\u064D\u0649])(?=(\W|$))', r'\1' + '\u0652', Strng)
    Strng = re.sub('(?<!\u0650)([\u064A])(?![\u0651\u0652\u064E\u0650\u064F\u064B\u064C\u064D\u0649])(?=(\W|$))', r'\1' + '\u0652', Strng)
    Strng = re.sub('(?<!\u0650)([\u064A]\u0651)(?![\u0652\u064E\u0650\u064F\u064B\u064C\u064D\u0649])(?=(\W|$))', r'\1' + '\u0652', Strng)
    Strng = re.sub('(?<!\u064F)([\u0648])(?![\u0651\u0652\u064E\u0650\u064F\u064B\u064C\u064D\u0649])(?=(\W|$))', r'\1' + '\u0652', Strng)
    Strng = re.sub('(?<!\u064F)([\u0648]\u0651)(?![\u0652\u064E\u0650\u064F\u064B\u064C\u064D\u0649])(?=(\W|$))', r'\1' + '\u0652', Strng)

    #print(Strng)
    #print('here2', Strng)

    return Strng

def AlephMaterLectionis(Strng, target='semitic'):
    Strng += "\u05CD"

    return Strng

# Indic Target
def FixSemiticRoman(Strng, Source):
    vir = '\u033D'

    if "\u05CD" in Strng:
        Strng = PostProcess.AlephMaterLectionis(Strng)

    if "\u05CC" in Strng:
        #print('before1', Strng)
        Strng = PostProcess.removeSemiticLetters(Strng)
        #print('after1', Strng)

        ## Fix Ayin -> Aleph approximation
        AyinAlephInitial = [('ʾa', 'ʾ'),('ʾā', 'ā̂'), ('ʾi', 'î'), ('ʾī', 'ī̂'), ('ʾu', 'û'), ('ʾū', 'ū̂'), ('ʾe', 'ê'), ('ʾē', 'ē̂'),\
                             ('ʾo', 'ô'), ('ʾō', 'ō̂')]

        for x, y in AyinAlephInitial:
            Strng = Strng.replace(x, y)

        ## Also make changes in PostOptionsLatn
        if Source == 'Arab':
            Strng = PostProcess.arabizeLatn(Strng, target="indic")
            #print('after2', Strng)
        elif Source == 'Arab-Ur' or Source == 'Arab-Pa' or Source == 'Arab-Fa':
            Strng = PostProcess.urduizeLatn(Strng, target="indic")
        elif Source == 'Syrn':
            Strng = PostProcess.syricizeLatn(Strng, target="indic")
        elif Source == 'Syrj' or Source == 'Hebr':
            Strng = PostProcess.hebraizeLatn(Strng, target="indic")

    # remove sin/shin dot
    Strng = Strng.replace('\u032A', '').replace('\u032E', '')

    # duplicate consonants
    ## move do chashmee ha + gemination sign
    ## tʰ꞉ -> t꞉ʰ
    Strng = re.sub('([ʰ])(꞉)', r'\2\1', Strng)

    Strng = re.sub('([aiuāīū])(꞉)', r'\2\1', Strng)
    Strng = re.sub('(.)(꞉)', r'\1' + vir + r'\1', Strng)

    # avoid fusion of Ayn
    Strng = Strng.replace("ʿ" + vir, "ʿ" + vir + "\u200B")

    cons_prev = '|'.join(GM.SemiticConsonants)

    if 'Syr' in Source: ## converting from Syriac through Latin
        consSyrc = "|".join(["ʾ", "b",  "v", "g", "ġ", "d", "ḏ", "h", "w", "z", "ḥ", "ṭ", "y", "k", "ḫ", "l", "m", "n", "s", "ʿ", "p", "f", "ṣ", "q", "r", "š", "t", "ṯ", "č", "ž", "j"])
        vowelSyrc = ["a", "ā", "e", "ē", "ū", "ō", "ī", "â", "ā̂", "ê", "ē̂"]

        vowelsDepSyrc = "|".join(["a", "ā", "e", "ē", "u", "i", "o"])
        vowelsInDepSyrc1 = ["i", "u", "o"]
        vowelsInDepSyrc2 = ["ī̂", "û", "ô"]

        if any([vow in Strng for vow in vowelSyrc]):
            #print(Strng)
            #print([(vow, vow in Strng) for vow in vowelSyrc])
            Strng = Strng.replace('ī', 'i').replace('ū', 'u').replace('ō', 'o')

            for vow1, vow2 in zip(vowelsInDepSyrc1, vowelsInDepSyrc2):
                Strng = re.sub('(?<!\w)' + vow1, vow2, Strng)

            Strng = Strng.replace('̂̂', '̂').replace('ô̂', 'ô') # [oi]^^ -> i^

            if "\uF001" in Strng:
                Strng = re.sub('(' + consSyrc + ')' + '(?!' + vowelsDepSyrc + ')', r'\1' + vir, Strng)

            # print(Strng)
            Strng = re.sub('(?<=' + cons_prev + ')' + 'a(?!\u0304)', '', Strng)

        Strng = Strng.replace("\uF001", "")

    ## Add scripts with vowels here
    if "Arab" in Source or Source == 'Latn' or Source == "Hebr" or Source == "Thaa" or Source == 'Type':
        basic_vowels = '(' + '|'.join(['a', 'ā', 'i', 'ī', 'u', 'ū', 'ē', 'ō', 'e', 'o', '#', vir]) + ')'
        Strng = re.sub('(ŵ)(?=' + basic_vowels + ')', "w", Strng)
        Strng = re.sub('(ŷ)(?=' + basic_vowels + ')', "y", Strng)

        Strng = re.sub('(?<=' + cons_prev + ')' + 'a(?!(ŵ|ŷ|\u0304|\u032E))', '', Strng)
        Strng = re.sub('(?<=ḧ)' + 'a(?!(ŵ|ŷ|\u0304|\u032E))', '', Strng)

        ## Hamza with vowels
        if 'Arab' in Source:
            simp_vow = 'a ā i ī u ū'.split(' ')
            init_vow = 'â ā̂ î ī̂ û ū̂'.split(' ')

            # Hamsa + vow -> indep vow
            for x, y in zip(simp_vow, init_vow):
                Strng = re.sub('ʔ' + x, y, Strng)

            # remove leftover hamza if nativizing
            if "\u05CC" in Strng:
                Strng = Strng.replace("ʔ", "")

    ## Semitic to Indic equivalents
    SemiticIndic=[('ṣ', 'sQ'), ('ʿ', 'ʾQ'), ('ṭ', 'tQ'), ('ḥ', 'hQ'), \
        ('ḍ', 'dQ'), ('p̣', 'pQ'), ('ž', 'šQ'), ('ẓ', 'jʰQ'), ('ḏ', 'dʰQ'), ('ṯ', 'tʰQ'),\
            ('w', 'vQ'), ('ḵ', 'k'), ('\u032A', ''), ('\u032E', ''),('a̮', "ā"),('\u0308', ""),\
                    ('ĕ\u0302', 'ê'), ('ă\u0302', 'â'), ('ŏ\u0302','ô'), ('ĕ', 'e'), ('ă', ''), ('ŏ','o'), ('ḵ', 'k'),\
                        ('ʾQā', 'ā̂Q'), ('ʾQi', 'îQ'), ('ʾQī', 'ī̂Q'), ('ʾQu', 'ûQ'), ('ʾQū', 'ū̂Q'), ('ʾQe', 'êQ'), ('ʾQē', 'ē̂Q'),\
                             ('ʾQo', 'ôQ'), ('ʾQō', 'ō̂Q'), ('ⁿ', 'n\u033D'), ('ʾā', 'ā̂')]

    for s, i in SemiticIndic:
        Strng = Strng.replace(s, i)

    if 'Arab' in Source:
        Strng = re.sub('(\u033D)([iuā])', r'\2', Strng)
        Strng = re.sub('(\u033D)([a])', '', Strng)

    ## Lone Aliph to Nukta
    Strng = Strng.replace('ʾ', 'â')

    return Strng

def perisanizeArab(Strng):
    arabKafYe = 'ك ي'.split(' ')
    persKafYe = 'ک ی'.split(' ')

    for x, y in zip(arabKafYe, persKafYe):
        Strng = Strng.replace(x, y)

    return Strng

def ArabizePersian(Strng):
    ## how to deal with ga
    arabKafYe = 'ك ي'.split(' ')
    persKafYe = 'ک ی'.split(' ')

    for x, y in zip(arabKafYe, persKafYe):
        Strng = Strng.replace(y, x)

    return Strng

def semiticizeUrdu(Strng):
    urduSpecific = 'ے ڈ ٹ ہ'.split(' ')
    semitic = 'ي د ت ه'.split(' ')

    for x, y in zip(urduSpecific, semitic):
        Strng = Strng.replace(x, y)

    ## remove Do chashme he
    Strng = Strng.replace('ھ', '')

    return Strng

def ShowChillus(Strng):

    return PostProcess.MalayalamChillu(Strng, True, True)

def ShowKhandaTa(Strng):
    Strng = Strng.replace('ৎ', 'ত্ˍ')

    return Strng

def eiaudipthongs(Strng):

    return Strng

def wasvnukta(Strng):

    return Strng

def default(Strng):

    return Strng

def SogdReshAyin(Strng):
    Strng = Strng.replace('𐽀', '[\uEA01-\uEA02]') # resh yin

    return Strng

def SogoReshAyinDaleth(Strng):
    Strng = Strng.replace('𐼘', '[\uEA01-\uEA02-\uEA06]') # resh yin daleth

    return Strng

def PhlpMemQoph(Strng):
    Strng = Strng.replace('𐮋', '[\uEA03-\uEA04]') # mem qoph

    return Strng

def PhlpWawAyinResh(Strng):
    Strng = Strng.replace('𐮅', '[\uEA05-\uEA02-\uEA01]') # resh yin

    return Strng

def PhliWawAyinResh(Strng):
    Strng = Strng.replace('𐭥', '[\uEA05-\uEA02-\uEA01]') # resh yin

    return Strng

def HatrDalethResh(Strng):
    Strng = Strng.replace('𐣣', '[\uEA06-\uEA01]') # resh yin

    return Strng

def MalayalamHalfu(Strng):
    consu = '[കചടതറനണരലവഴളറ]'
    vir = GM.CrunchSymbols(GM.VowelSigns, 'Malayalam')[0]
    consAll = "(" + '|'.join(GM.CrunchSymbols(GM.Consonants, 'Malayalam')) + ")"

    Strng = re.sub('(?<=' + consu + ')' + '(' + vir + ')' + '(?!' +  consAll + ')', r'\2' + 'ു്', Strng)

    return Strng

def MalayalamTranscribe(Strng):
    Strng = MalayalamHalfu(Strng)

    script = "Malayalam"

    ListC = GM.CrunchList('ConsonantMap',script)
    ListSC = GM.CrunchList('SouthConsonantMap',script)
    vir = GM.CrunchSymbols(GM.VowelSigns, script)[0]

    ConUnVoiced = [ListC[x] for x in [0,5,10,15,20]]
    ConVoicedJ =  [ListC[x] for x in [2,7,12,17,22]]
    ConVoicedS =  [ListC[x] for x in [2,5,12,17,22]]

    ConNasalsAll = '|'.join([ListC[x] for x in [4, 9, 14, 19, 24]])
    conNasalCa = '|'.join([ListC[x] for x in [9]])
    ConNasalsGroup = [ConNasalsAll, conNasalCa, ConNasalsAll, ConNasalsAll, ConNasalsAll]

    ConMedials = '|'.join(ListC[25:28]+ListSC[0:2]+ListSC[3:4])
    Vowels = '|'.join(GM.CrunchSymbols(GM.Vowels+GM.VowelSignsNV, script))
    Aytham = GM.CrunchList('Aytham',script)[0]
    Consonants = '|'.join(GM.CrunchSymbols(GM.Consonants,script))

    NRA =ListSC[3] + vir + ListSC[2]
    NDRA = ListC[14] + vir + ListC[12] + vir + ListC[26]

    ### Check Siva Siva Mails
    ### Do something about Eyelash ra in Transliterated text

    for i in range(len(ConUnVoiced)):
        pass
        #Strng = re.sub('('+Vowels+Consonants+')'+ConUnVoiced[i]+'('+Vowels+Consonants+')',r'\1'+ConVoicedS[i]+r'\2',Strng)
        Strng = re.sub('('+Vowels+'|'+Consonants+'|'+Aytham+')'+ConUnVoiced[i]+'(?!'+vir+')',r'\1'+ConVoicedS[i],Strng)
        Strng = re.sub('('+ConVoicedS[i]+')'+ConUnVoiced[i]+'(?!'+vir+')',r'\1'+ConVoicedS[i],Strng)

        # Nasals + Unvoiced -> voiced
        # Rule applied even if spaced -> ulakaJ cuRRu -> ulakaJ juRRu; cenJu -> senju
        # Strng = re.sub('('+ConNasalsSpace+')'+'('+vir+')'+'( ?)'+ConUnVoiced[i],r'\1\2\3'+ConVoicedJ[i],Strng)

        # Only without space : vampu -> vambu;  varim panam -> varim panam
        Strng = re.sub('('+ConNasalsGroup[i]+')'+'('+vir+')'+ConUnVoiced[i],r'\1\2'+ConVoicedJ[i],Strng)

        # Medials + Unvoiced -> Voiced
        Strng = re.sub('('+ConMedials+')'+'('+vir+')'+ConUnVoiced[i]+'(?!'+vir+')',r'\1\2'+ConVoicedS[i],Strng)

    #RRA NRA

    Strng = Strng.replace('റ്റ', 'ട്ട').replace('ന്റ', 'ണ്ഡ')

    #anusvara with m
    Strng = Strng.replace('ം', 'മ്')

    return Strng

def retainLatin(Strng, reverse=False):
    latn_basic_lower = 'a b c d e f g h i j k l m n o p q r s t u v w x y z ḥ ṭ ṣ ʾ ʿ š ā ī ū ē ō'
    latn_basic_upper = latn_basic_lower.upper()
    latn_all = latn_basic_lower + latn_basic_upper
    latn_all = latn_all.split(' ')

    if not reverse:
        for i, c in enumerate(latn_all):
            Strng = Strng.replace(c, chr(60929+i))

        # for disambiguating preprocessed consonants
        Strng = Strng.replace('\uEA01', 'r').replace('\uEA02', 'ʿ').replace('\uEA03', 'm').replace('\uEA04', 'q').replace('\uEA05', 'w').replace('\uEA06', 'd')

    else:
        # print('I am trying to reverse Lating')
        for i, c in enumerate(latn_all):
            Strng = Strng.replace(chr(60929+i), c)

    return Strng

def JapanesePreProcess(src, txt, preoptions):
    import pykakasi
    from . import Convert

    if src == 'Hiragana' or src == 'Katakana':
        # preserve roman words
        kks = pykakasi.kakasi()

        txt = Convert.convertScript(txt.lower(), "ISO", "Devanagari")
        #print(txt)

        cv = kks.convert(txt)
        txt = ''

        for item in cv:
            txt = txt + ' ' + item['hepburn']

        if 'eiaudipthongs' in preoptions:
            txt = txt.replace('ou', 'o\u02BDu').replace('ei', 'e\u02BDi')

        txt = re.sub('(r)([aiueo])(\u309A\u309A)', 'l' + r'\2\2', txt)
        txt = re.sub('(r)([aāiīuūeēoō])(\u309A)', 'l' + r'\2', txt)

        txt = re.sub('(k)([aiueo])(\u309A\u309A)', 'ṅ' + r'\2\2', txt)
        txt = re.sub('(k)([aāiīuūeēoō])(\u309A)', 'ṅ' + r'\2', txt)

        txt = txt.replace('aa', 'ā').replace('ii', 'ī').replace('ee', 'ē').replace('oo', 'ō').replace('uu','ū')
        txt = txt.replace('a-', 'ā').replace('i-', 'ī').replace('e-', 'ē').replace('o-', 'ō').replace('u-','ū')
        txt = txt.replace("n'", 'n_').replace('ch', 'c').replace('sh', 'ṣ').replace('sṣ', 'ṣṣ').replace('ai', 'a_i').replace('au', 'a_u')
        txt = txt.replace('w', 'v')
        txt = txt.replace('ou', 'ō').replace('ei', 'ē')
        txt = txt.replace('、', ',').replace('。', '.')

        # Nasalization of n

        txt= txt.replace('ng', 'ṅg').replace('nk', 'ṅk').replace('nk', 'ṅk').replace('np', 'mp').replace('nb', 'mb').replace('nm', 'mm')

        if 'wasvnukta' in preoptions:
            txt = txt.replace('v', 'v̈')

        #print(txt)

        txt = txt.replace('、', ',').replace('。', '.')

    return txt

def holamlong(Strng):
    Strng = Strng.replace('ֹּ','ֹּ')
    Strng = re.sub('(?<!ו)ֹ', 'וֹ', Strng)

    return Strng

def novowelshebrewIndic(Strng):
    Strng = novowelshebrewSemitic(Strng)

    finals = ['ך', 'ם', 'ן', 'ף', 'ץ', 'ףּ', 'ךּ']
    otherCons = 'ב,ח,ע,צ,ש,ת'.split(',')
    consonantsAll = '(' + '|'.join(GM.CrunchSymbols(GM.Consonants, 'Hebrew') + finals  + otherCons + ['׳', 'י', 'ו' ,'א']) + ')'
    vowelsignsADShinG = '(' + '|'.join(GM.CrunchSymbols(GM.VowelSigns, 'Hebrew') + ['ַ', 'ּ', 'ׁ', '׳']) + ')'

    Strng = re.sub(consonantsAll + "(?!" + vowelsignsADShinG + ')', r'\1' + 'ַ' + r'\2', Strng)

    return Strng

def novowelshebrewSemitic(Strng):
    Strng = Strng.replace('כ', 'כּ').replace('פ', 'פּ').replace( 'ב', 'בּ')
    Strng = Strng.replace('ך', 'ךּ').replace('ף','ףּ')

    return Strng

def shvanakhall(Strng):
    Strng = Strng + ' \u0BDE'

    return Strng

def longEOISO(Strng):
    Strng = Strng.replace('e', 'ē').replace('o', 'ō')

    return Strng

def SanskritLexicaizeHK(Strng):

    return Strng

def ThaiPhonetic(Strng):
    Strng = Strng.replace('ด', 'ท')
    Strng = Strng.replace('บ', 'พ')
    Strng = Strng.replace('ก\u0325', 'ค')
    Strng = Strng.replace('จ\u0325', 'ช')
    Strng = Strng.replace('งํ', 'ง')

    Strng = Strng.replace('\u035C', '')

    Strng = Strng.replace('\u0E47', '')

    Strng += "\u02BB\u02BB"

    return Strng

def LaoPhonetic(Strng):
    Strng = Strng.replace('ດ', 'ທ')
    Strng = Strng.replace('ບ', 'ພ')
    Strng = Strng.replace('ງໍ', 'ງ')

    Strng = Strng.replace('\u035C', '')

    Strng += "\u02BB\u02BB"

    return Strng

def SaurastraHaaruColonTamil(Strng):
    Strng = Strng.replace('ன', 'ந')

    ListVS = '|'.join(GM.CrunchSymbols(GM.VowelSigns, 'Tamil'))

    Strng = re.sub('(' + ListVS + ')' + '(:)' , r'\2\1', Strng)

    chars = '([நமரல])'

    Strng = re.sub(chars + ':', r'\1' + '\uA8B4', Strng)

    return Strng

def ChakmaPali(Strng):
    Strng = Strng.replace('\U00011147', '𑄤') # Replace Ya
    Strng = Strng.replace('𑄠', '𑄡') # Replace vA

    listC = '('+"|".join(sorted(GM.CrunchSymbols(GM.Consonants,"Chakma")+Chakma.VowelMap[:1],key=len,reverse=True))+')'
    listV = '('+"|".join(sorted(GM.CrunchSymbols(GM.VowelSigns,"Chakma")+Chakma.ViramaMap+['\U00011133'],key=len,reverse=True))+')'

    Strng = Strng.replace("\u02BD","")

    Strng = Strng.replace('\U00011102', '\U00011127')

    # Introduce vowel Sign A ; Chakma - Inharant vowel is AA
    Strng = re.sub("("+listC+")"+"(?!"+listV+")",r'\1''\u02BE',Strng)
    Strng = Strng.replace("\U00011127","")
    Strng = Strng.replace("\u02BE","\U00011127")

    return Strng

def TakriArchaicKha(Strng):

    return Strng.replace('𑚋', '𑚸')

def UrduShortNotShown(Strng):
    Strng += "\u02BB\u02BB"

    return Strng

def AnuChandraEqDeva(Strng):

    return AnuChandraEq(Strng, 'Devanagari')

def AnuChandraEq(Strng, script):
    Chandrabindu = GM.CrunchList('AyogavahaMap', script)[0]
    Anusvara = GM.CrunchList('AyogavahaMap', script)[1]

    Strng = Strng.replace(Chandrabindu, Anusvara)

    return Strng

def TamilNumeralSub(Strng):
    ListC = '(' + '[கசடதபஜஸ]' + ')'
    ListV = '(' + '|'.join(GM.CrunchSymbols(GM.VowelSigns, 'Tamil')) + ')'

    Strng = re.sub(ListC + ListV + '2', r'\1\2' + '²', Strng)
    Strng = re.sub(ListC + ListV + '3', r'\1\2' + '³', Strng)
    Strng = re.sub(ListC + ListV + '4', r'\1\2' + '⁴', Strng)

    Strng = re.sub(ListC + '2', r'\1' + '²', Strng)
    Strng = re.sub(ListC + '3', r'\1' + '³', Strng)
    Strng = re.sub(ListC + '4', r'\1' + '⁴', Strng)

    Strng = Strng.replace('ரு\'', 'ருʼ')
    Strng = Strng.replace('ரு’', 'ருʼ')

    Strng = Strng.replace('ம்\'', 'ம்ʼ')
    Strng = Strng.replace('ம்’', 'ம்ʼ')

    return Strng

def swapEe(Strng):
    Strng = Strng.replace('E', 'X@X@')
    Strng = Strng.replace('e', 'E')
    Strng = Strng.replace('X@X@','e')

    Strng = Strng.replace('O', 'X@X@')
    Strng = Strng.replace('o', 'O')
    Strng = Strng.replace('X@X@','o')

    return Strng

def swapEeItrans(Strng):
    Strng = Strng.replace('^e', 'X@X@')
    Strng = Strng.replace('e', '^e')
    Strng = Strng.replace('X@X@','e')

    Strng = Strng.replace('^o', 'X@X@')
    Strng = Strng.replace('o', '^o')
    Strng = Strng.replace('X@X@','o')

    return Strng

def egrantamil(Strng):
    return Strng

def siddhammukta(Strng):
    return Strng

def TaiKuen(Strng):
    return Strng

def TaiThamLao(Strng):
    return Strng

def ThaiSajjhayaOrthography(Strng):
    Script = "Thai"

    #cons = "|".join(GM.CrunchSymbols(GM.Consonants, Script)+GM.CrunchList('VowelMap',Script)[0:1])
    #EAIO = "".join(GM.CrunchList('VowelSignMap',Script)[9:12]+GM.CrunchList('SinhalaVowelSignMap',Script)[:])
    ## Reorder dve
    #Strng = re.sub('([' + EAIO + '])' + '(' + cons  + ')' + '(๎)' + '(' + cons + ')', r'\2\3\1\4', Strng)

    Strng = Strng.replace('ัง', 'ังฺ')
    Strng = Strng.replace('์', 'ฺ')
    Strng = Strng.replace('๎', 'ฺ')
    Strng = Strng.replace('ั', '')

    return Strng

def ThaiSajjhayawithA(Strng):
    Strng = Strng.replace('ะ', '')
    Strng = ThaiSajjhayaOrthography(Strng)

    return Strng

def LaoSajhayaOrthography(Strng):
    Strng = Strng.replace('ັງ', 'ັງ຺')

    Strng = re.sub('([ເໂໄ])(.๎)([ຍຣລວຨຩສຫຬ])', r'\2\1\3', Strng)

    Strng = Strng.replace('໌', '຺')
    Strng = Strng.replace('๎', '຺')
    Strng = Strng.replace('ັ', '')

    return Strng

def LaoSajhayaOrthographywithA(Strng):
    Strng = Strng.replace('ະ', '')
    Strng = LaoSajhayaOrthography(Strng)

    return Strng

# consider adding an optional NUkta to the post consonantal position
def RemoveSchwaHindi(Strng, showschwa=False):
    VowI = "(" + '|'.join(GM.CrunchSymbols(GM.Vowels,'Devanagari')) + ")"
    VowS = "(" + '|'.join(GM.CrunchSymbols(GM.VowelSignsNV, 'Devanagari')) + ")"
    Cons = "(" + '|'.join(GM.CrunchSymbols(GM.Consonants, 'Devanagari')) + ")"
    Char = "(" + '|'.join(GM.CrunchSymbols(GM.Characters, 'Devanagari')) + ")"
    Nas = "([ंःँ]?)"
    ISyl = "((" + VowI + "|" + "(" + Cons + VowS + "?" + "))" + Nas +')'
    Syl = "((" + Cons + VowS + ')' + Nas + ")"
    SylAny = "((" + Cons + VowS + "?" + ')' + Nas + ")"

    if not showschwa:
        vir = '्'
        vir2 = '्'
    else:
        vir = '\u0954'
        vir2 = '\u0954'

    Strng = re.sub(ISyl+Cons+Cons+SylAny+"(?!" + Char + ")", r'\1\8' + vir + r'\9\10', Strng) # bhAratIya --> bhArtIy
    Strng = re.sub(ISyl+Cons+Syl+SylAny+"(?!" + Char + ")", r'\1\8' + vir + r'\9\15', Strng) # bhAratIya --> bhArtIy
    Strng = re.sub(ISyl+Cons+Syl+"(?!" + Char + ")", r'\1\8' + vir + r'\9', Strng) # namakIn -> namkIn

    Strng = re.sub(ISyl+Cons+"(?!" + Char + ")", r'\1\8' + vir, Strng) # kama -> kam

    Cons_sss = '((' + Cons + vir + ')' +'([शषस]))' # delete if double ends with s, sh, s
    Strng = re.sub(ISyl+Cons_sss+"(?!" + Char + ")", r'\1\8' + vir, Strng)

    ## remove from geminated consonants
    Target = 'Devanagari'

    ConUnAsp = [GM.CrunchList('ConsonantMap', Target)[x] for x in [0,2,5,7,10,12,15,17,20,22,4,9,14,19,24,25,26,27,28,29,30,31,32]]
    ConUnAsp = ConUnAsp + GM.CrunchList('SouthConsonantMap',Target) + GM.CrunchList('NuktaConsonantMap',Target)
    ConAsp   = [GM.CrunchList('ConsonantMap', Target)[x] for x in [1,3,6,8,11,13,16,18,21,23]]

    Strng = re.sub(ISyl + '('+'|'.join(ConUnAsp)+')'+'('+vir+')('+r'\8'+")(?!" + Char + ")", r'\1\8\9\10' + vir,Strng)

    for i in range(len(ConAsp)):
        Strng = re.sub(ISyl +'('+ConUnAsp[i]+')'+'('+vir+')'+'('+ConAsp[i]+')'+'(?!" + Char + ")', r'\1\8\9\10' + vir,Strng)

    #Strng = re.sub(VowI + Nas + Cons2+"(?!" + Char + ")", r'\1\2\3' + vir, Strng)

    cons_pyramid = ['[यरलव]', '[नमण]', '[शषस]', '[कखपफगघबभ]', '[टठतथडढदध]', '[चछजझज़]']
    for c1, cons1 in enumerate(cons_pyramid):
        for c2, cons2 in enumerate(cons_pyramid):
            if c1 < c2: # delete if ascending order
                Cons_pyr = '((' + cons1 + vir + ')' +'('+cons2+'))'
                Strng = re.sub(ISyl+Cons_pyr+"(?!" + Char + ")", r'\1\8' + vir, Strng)


    Strng = Strng.replace(vir, vir2)

    return Strng

# consider adding an optional NUkta to the post consonantal position
def RemoveFinal(Strng, Target):
    if Target == 'Bengali':
        Strng = PostProcess.KhandaTa(Strng, Target, True)

    VowI = "(" + '|'.join(GM.CrunchSymbols(GM.Vowels,Target)) + ")"
    VowS = "(" + '|'.join(GM.CrunchSymbols(GM.VowelSignsNV, Target)) + ")"
    Cons = "(" + '|'.join(GM.CrunchSymbols(GM.Consonants, Target)) + ")"
    Char = "(" + '|'.join(GM.CrunchSymbols(GM.Characters, Target)) + ")"
    Nas = "([" + '|'.join(GM.CrunchList('AyogavahaMap',Target)) + "]?)"

    ISyl = "((" + VowI + "|" + "(" + Cons + VowS + "?" + ")" + Nas + '))'
    Syl = "((" + Cons + VowS + ')' + Nas + ")"
    SylAny = "((" + Cons + VowS + "?" + ')' + Nas + ")"

    vir = GM.CrunchList("ViramaMap", Target)[0]
    if Target != 'Bengali':
        Cons2 = '((' + Cons + vir + ')?' + Cons + ')'
    else:
        Cons2 = '(()?' + Cons + ')'

    if Target == 'Khmer' or Target == 'KhmerLoC':
        ListC ='|'.join(GM.CrunchSymbols(sorted(GM.Consonants, key=len, reverse=True),Target))
        ra = Khmer.ConsonantMap[26]
        Strng = re.sub('('+ListC+')'+'\u17CC',ra+'\u17D2'+r'\1',Strng)

        Strng = re.sub(ISyl + '('+ListC +')' + '(((\u17D2)' + '('+ListC +'))*)([៍៎៏]?)(?=[\s\n])', r'\1\8\9' + vir + r'\13', Strng) # kama -> kam
        Strng = re.sub(ISyl + '('+ListC +')' + '(((\u17D2)' + '('+ListC +'))*)([៍៎៏]?)$', r'\1\8\9' + vir + r'\13', Strng) # kama -> kam
    else:
        Strng = re.sub(ISyl + Cons2+"(?!" + Char + ")", r'\1\8' + vir, Strng) # kama -> kam
        Strng = re.sub(ISyl + Cons2+"(?!" + Char + ")", r'\1\8' + vir, Strng) # kama -> kam

    #Strng = re.sub(VowI + Nas + Cons2+"(?!" + Char + ")", r'\1\2\3' + vir, Strng)

    return Strng

def SchwaFinalGurmukhi(Strng):

    Strng = RemoveFinal(Strng, 'Gurmukhi')

    return Strng

def SchwaFinalGujarati(Strng):

    Strng = RemoveFinal(Strng, 'Gujarati')

    return Strng

def SchwaFinalBengali(Strng):

    Strng = RemoveFinal(Strng, 'Bengali')

    return Strng

def SchwaFinalKhmerLoC(Strng):

    Strng = RemoveFinal(Strng, 'KhmerLoC')

    return Strng

def SchwaFinalKhmer(Strng):

    Strng = RemoveFinal(Strng, 'Khmer')

    return Strng

def SchwaFinalWarangCiti(Strng):
    Target = "WarangCiti"

    VowI = "(" + '|'.join(GM.CrunchSymbols(GM.Vowels,Target)) + ")"
    VowS = "(" + '|'.join(GM.CrunchSymbols(GM.VowelSignsNV, Target)) + ")"
    Cons = "(" + '|'.join(GM.CrunchSymbols(GM.Consonants, Target)) + ")"
    Char = "(" + '|'.join(GM.CrunchSymbols(GM.Characters, Target)) + ")"

    Nas = "([" + '|'.join(GM.CrunchList('AyogavahaMap',Target)) + "]?)"

    ISyl = "((" + VowI + "|" + "(" + Cons + VowS + "?" + ")" + Nas + '))'
    Syl = "((" + Cons + VowS + ')' + Nas + ")"
    SylAny = "((" + Cons + VowS + "?" + ')' + Nas + ")"

    vir = '\u02BB'
    Cons2 = '((' + Cons + vir + ')?' + Cons + ')'

    Strng = re.sub(ISyl+Cons2+"(?!" + Char + ")", r'\1\8' + vir, Strng) # kama -> kam

    return Strng

def siddhamUnicode(Strng):
    return Strng

def ThaiOrthography(Strng):
    Strng += "\u02BB\u02BB"

    return Strng

def LaoTranscription(Strng):
    Strng += "\u02BB\u02BB"

    return Strng

def LimbuDevanagariConvention(Strng):
    Strng = Strng.replace('ए़', 'ऎ')
    Strng = Strng.replace('ओ़', 'ऒ')
    Strng = Strng.replace('े़', 'ॆ')
    Strng = Strng.replace('ो़', 'ॊ')
    Strng = Strng.replace('ः', '꞉')

    return Strng

def LimbuSpellingSaI(Strng): ## Fix this not for all
    vir = Limbu.ViramaMap[0]

    FCons = [x+vir for x in [Limbu.ConsonantMap[x] for x in[0,4,15,19,20,24,26,27]]]
    FinalCons = ['\u1930','\u1931','\u1933','\u1934','\u1935','\u1936','\u1937','\u1938']

    for x, y in zip(FCons, FinalCons):
        Strng = Strng.replace(x, '\u193A' + y)

    return Strng


def removeChillus(Strng):
    Chillus=['\u0D7A','\u0D7B','\u0D7C','\u0D7D','\u0D7E']

    vir = Malayalam.ViramaMap[0]
    ConVir =[
             Malayalam.ConsonantMap[14]+vir,
             Malayalam.ConsonantMap[19]+vir,
             Malayalam.ConsonantMap[26]+vir,
             Malayalam.ConsonantMap[27]+vir,
             Malayalam.SouthConsonantMap[0]+vir,
            ]

    for x,y in zip(Chillus, ConVir):
        Strng = Strng.replace(x, y)

    return Strng

def SinhalaPali(Strng):
    Strng = PostProcess.SinhalaPali(Strng, reverse = True)

    return Strng

def IASTPali(Strng):
    Strng = Strng.replace('ḷ', 'l̤')

    return Strng

def CyrillicPali(Strng):
    Strng = Strng.replace('л̣', 'л̤',)

    return Strng

def MalayalamPrakrit(Strng):
    ## Reverse Gemination
    Strng = PostProcess.ReverseGeminationSign(Strng, 'Malayalam')
    Strng = Strng.replace("ഀ", "ം")

    return Strng

def GranthaPrakrit(Strng):
    ## Reverse Gemination
    Strng = PostProcess.ReverseGeminationSign(Strng, 'Grantha')

    Strng = Strng.replace("𑌀", "𑌂")

    return Strng

def RomanPreFix(Strng,Source):
    DepV = '\u1E7F'
    Asp = '\u02B0'

    Vir = GM.CrunchList("ViramaMap", Source)[0]
    Nuk = GM.CrunchList("NuktaMap", Source)[0]
    VowelA = GM.CrunchSymbols(['VowelMap'],Source)[0]

    ListV = '|'.join(GM.CrunchSymbols(GM.VowelSigns,Source))
    ListC = '|'.join(GM.CrunchSymbols(GM.Consonants,Source))

    Strng = re.sub('('+ListC+')'+'(?!'+ListV+'|'+VowelA+')',r'\1'+DepV+Vir,Strng)

    Strng = re.sub('('+ListC+'|'+Nuk+')'+'('+ListV+')',r'\1'+DepV+r'\2',Strng)
    #print Strng

    Strng = re.sub('(?<='+ListC+')'+'('+VowelA+')',r'',Strng)
    #print Strng

    #Fix Nukta त़् त़्अ त़्ा -->  त़् त़ त़ा
    Strng = Strng.replace(DepV+Vir+Nuk+VowelA,Nuk)
    Strng = re.sub(DepV+Vir+Nuk+'(?=['+DepV+'])',Nuk,Strng)
    Strng = Strng.replace(DepV+Vir+Nuk,Nuk+DepV+Vir)

    #print Strng

    return Strng

def joinVowelCons(Strng, script):
    consonantsAll = '(' + '|'.join(sorted(GM.CrunchSymbols(GM.Consonants, script), key = len, reverse=True)) + ')'
    vowelsAll = '(' + '|'.join(sorted(GM.CrunchSymbols(GM.Vowels, script), key = len, reverse=True)) + ')'

    Strng = re.sub(consonantsAll + ' ' + vowelsAll, r'\1\2', Strng)
    Strng = re.sub(consonantsAll + ' ' + consonantsAll, r'\1\2', Strng)

    return Strng

def joinVowelConsIAST(Strng):
    return joinVowelCons(Strng, 'IAST')

def joinVowelConsISO(Strng):
    return joinVowelCons(Strng, 'ISO')

def PreProcess(Strng,Source,Target,postoptions,preoptions):
    if Source in GM.RomanDiacritic or Source == 'Latn':
        Strng = Strng.lower()

    if Source in GM.pipeScripts:
        if '|' in Strng:
            Strng = Strng.replace(".", "●")
            Strng = Strng.replace("।", "|").replace("॥", "||")
            Strng = Strng.replace("|", ".").replace("||", "..")
        if Source == "Itrans":
            Strng = re.sub("●([acCDN])", "."+r'\1', Strng)
        elif Source == "Velthuis":
            Strng = re.sub("●([rRlLmhtdnTDsyg])", "."+r'\1', Strng)
            Strng = re.sub("●●([lr])", "."+r'\1', Strng)

    if 'Arab' in Source:
        Strng = re.sub('([وي])(?=[\u064E\u0650\u064F\u0651\u064B\u064C\u064D])', '\u02DE' + r'\1', Strng)
        #Strng = re.sub('((\u064E|\u0650|\u0652|\u064F))(\u0651)', r'\2', Strng)

    if Source in ["Syrj", "Syrn"]:
        Strng = Strng.replace('\u0323', '\u0742')


    if Source == 'Itrans':
        sOm = 'OM'
        tOm = 'oM'

        punc =  '(' + '|'.join(["\u005C"+x for x in list(string.punctuation)]+ ['\s']
                    + [x.replace('.', '\.') for x in GM.CrunchSymbols(GM.Signs,Source)[1:3]]) + ')'

        Strng = re.sub(punc + sOm + punc, r'\1' + tOm + r'\2', Strng)
        Strng = re.sub('^' + sOm + punc, tOm + r'\1', Strng)
        Strng = re.sub(punc + sOm + '$', r'\1' + tOm, Strng)
        Strng = re.sub('^' + sOm + '$', tOm, Strng)

        punc = '(\s)'

        Strng = re.sub(punc + sOm + punc, r'\1' + tOm + r'\2', Strng)
        Strng = re.sub('^' + sOm + punc, tOm + r'\1', Strng)
        Strng = re.sub(punc + sOm + '$', r'\1' + tOm, Strng)
        Strng = re.sub('^' + sOm + '$', tOm, Strng)


        # kaRRi -> katri etc. not kaVoc.R
        Strng = re.sub('([aAiuUeoEO])([R]){2}([iI])', r'\1'+r'\2'+'_'+r'\2'+r'\3', Strng)
        Strng = re.sub('([aAiuUeoEO])([L]){2}([iI])', r'\1'+r'\2'+'_'+r'\2'+r'\3', Strng)

        AltForm = ['O', 'aa','ii','uu','RRi','RRI','LLi','LLI','N^','JN','chh','shh','x','GY','.n','.m','.h', 'AUM', 'E', 'J', 'c.o', 'c.e']
        NormForm = ['^o', 'A','I','U','R^i','R^I','L^i','L^I','~N','~n','Ch','Sh','kSh','j~n','M','M','','oM', '^e', 'z', 'A.c', 'e.c']

        for x,y in zip(AltForm,NormForm):
            Strng = Strng.replace(x,y)

        AltForms = [('ee', 'I'), ('dny', 'j~n'), ('oo', 'U'), ('kS', 'kSh'), ('w', 'v'), ('|', '.'), ('kShh', 'kSh')]

        for x,y in AltForms:
            Strng = Strng.replace(x,y)

        Strng = Strng.replace('OM', 'oM')


    if Source == 'BarahaNorth' or Source == 'BarahaSouth':
        # alternate representations

        alt_baraha = [('A', 'aa'), ('I', 'ee'), ('U', 'oo'), ('~loo', '~lU'), ('Roo', 'RU'), ('ou', 'au'), ('K', 'kh'), ('G','gh'), ('ch', 'c'), ('Ch', 'C'), ('J','jh'), ('P', 'ph'), ('B', 'bh'), ('w', 'v'), ('sh', 'S'), ('~h', '_h'), ('Y', 'yx'), ('^^', '{}'), ('^', '()'), ('tx', 'rx'), ('zh', 'Lx'), ('~e', '~a'), ('q', '\_'), ('#', "\\'"), ('$', '\\"')]

        for alt, norm in alt_baraha:
            Strng = Strng.replace(alt, norm)

        if Target == 'Tamil':
            Strng = Strng.replace('n', 'nx').replace('~nx', 'n')
            print(Strng)

    if 'IAST' in Source:
        Strng = Strng.replace("aï", "a_i")
        Strng = Strng.replace("aü", "a_u")
        Strng = Strng.replace('\u0303', 'ṃ')

    if "ISO" in Source:
        Strng = Strng.replace('a:i', 'a_i')
        Strng = Strng.replace('a:u', 'a_u')
        Strng = Strng.replace('\u0303', 'ṁ')
        ## People use the wrong convention sometimes
        Strng = Strng.replace('ṃ', 'ṁ')

    if Source == "Titus":
        Strng = Strng

    if Source == "ISO" or Source == "IAST" or Source == "Titus" or "RussianCyrillic":
        Strng = CF.VedicSvarasNonDiacritic(Strng)

    if Source == "Latn" and 'Syr' in Target:
        Strng = Strng.replace('ḇ', 'v').replace('ḡ', 'ḡ').replace('ḵ', 'ḫ').replace('p̄', 'f')

    if ('↓' in Strng or '↑' in Strng) and Target in GM.IndicScripts :
        Strng = Strng.replace('↓', '॒')
        Strng = Strng.replace('↑↑', '᳚')
        Strng = Strng.replace('↑', '॑')

    if ('↓' in Strng or '↑' in Strng) and Target in GM.LatinScripts :
        Strng = Strng.replace('↓', '\\_')
        Strng = Strng.replace('↑↑', '\\"')
        Strng = Strng.replace('↑', '\\\'')

    if Source == "WarangCiti":
        Strng = Strng.replace('\u200D', '\u00D7')

    if Source == "Hebr-Ar":
        # accept variants with dotes
        dot_var = [('עׄ','ג'), ('תׄ','ת֒'), ('ת','ת̈'), ('ק','ק̈')]

        for char, char_var in dot_var:
            Strng = Strng.replace(char_var, char)

    ## Normalize Input Strings

    Strng = normalize(Strng,Source)

    ## Remove unsupported letters and replace with supported ones

    #Fix Udata + Avagraha combination
    # नमो॑ऽयम् --> namo̍​'yam
    if Source in GM.IndicScripts and Target in GM.Transliteration:
        udatta = '\u0951'
        avagrahaSrc = GM.CrunchList('SignMap', Source)[0]
        avagrahaTgt = GM.CrunchList('SignMap', Target)[0]

        if avagrahaTgt == "'":
            Strng = Strng.replace(udatta + avagrahaSrc, udatta + '\u200B' + avagrahaSrc)

    return Strng

def ISO259Target(Strng):
    Strng = Strng.replace('א', 'ʾ').replace('׳', '’')

    return Strng

def ISO233Target(Strng):
    replacements = [('أ', 'ˈʾ'), ('ء', '¦'), ('إ', 'ˌʾ')]

    for x, y in replacements:
        Strng = Strng.replace(x, y)

    return Strng

def PersianDMGTarget(Strng):
    replacements = [('ا', 'ʾ')]

    for x, y in replacements:
        Strng = Strng.replace(x, y)

    return Strng

def ISO233Source(Strng):
    replacements = [('أ', 'ˈʾ'), ('ء', '¦'), ('إ', 'ˌʾ')]

    for x, y in replacements:
        Strng = Strng.replace(y, x)

    replacements = [('j', 'ǧ'), ('g', 'ǧ'), ('ḧ', 'ẗ'), ('ḫ', 'ẖ'), ('a̮', 'ỳ'), \
        ('aⁿ', 'á'), ('iⁿ', 'í'), ('uⁿ', 'ú'), ('ā̂', 'ʾâ'), ('ˀ', 'ˈ')]

    for x, y in replacements:
        Strng = Strng.replace(y, x)

    return Strng

def HebrewSBLTarget(Strng):
    Strng = Strng.replace('א', 'ʾ').replace('׳', '’')

    return Strng

def HebrewSBLSource(Strng):
    Strng = Strng.replace('ʾ', 'א',).replace('’', '׳')
    Strng = Strng.replace('\u0307\u00B0', '\u00B0\u0307')

    replacements = [('v', 'ḇ'), ('f', 'p̄'), ('d꞉', 'd'), ('d', 'ḏ'), \
        ('g꞉', 'g'), ('g', 'ḡ'), \
            ('t꞉', 't'), ('t', 'ṯ'),
         ('š̮', 'š'), ('š̪', 'ś'),
        ('o', 'ō'), ('ō', 'ô'), ('ū', 'û'), ('\u033D', 'ĕ')
    ]

    for x, y in replacements:
        Strng = Strng.replace(y, x)

    return Strng

def ISO259Source(Strng):
    Strng = Strng.replace('ʾ', 'א',).replace('’', '׳')
    Strng = Strng.replace('\u0307\u00B0', '\u00B0\u0307')

    replacements = [('ḵ', 'k'), ('v', 'b'), ('f', 'p'), ('b', 'ḃ'), ('p', 'ṗ'), ('k', 'k̇'), ('꞉', '\u0307'),\
        ('š̮', 'š'), ('š', 's̀'), ('š̪', 'ś'),
        ('ā', 'å'), ('e', 'ȩ'), ('ō', 'ŵ'), ('ū', 'ẇ'), ('\u033D', '°'), ('ĕ', 'ḝ')
    ]

    for x, y in replacements:
        Strng = Strng.replace(y, x)

    Strng = unicodedata.normalize('NFD', Strng)
    Strng = Strng.replace('\u0307', '꞉')
    Strng = unicodedata.normalize('NFC', Strng)

    return Strng

def UnSupThaana(Strng):

    return Strng

def RemoveJoiners(Strng):
    Strng = Strng.replace("\u200D", "")
    Strng = Strng.replace("\u200C", "")

    return Strng

def ArabicGimelJa(Strng):
    Strng = Strng.replace("ج", "ڨ")

    return Strng

# Normalize Input
# ka + <nukta> -> qa etc
def normalize(Strng,Source):

    # Replace Decomposed Nukta Characters by pre-composed Nukta consonants

    #nuktaDecom = [u"क़",u"ख़",u"ग़",u"ज़",u"ड़",u"ढ़",u"फ़",u"य़",u"ਲ਼",u"ਸ਼",u"ਖ਼",u"ਗ਼",u"ਜ਼",u"ਫ਼",u"ড়",u"ঢ়",u"য়",u"ଡ଼",u"ଢ଼"]
    #nuktaPrecom = [u"क़",u"ख़",u"ग़",u"ज़",u"ड़",u"ढ़",u"फ़",u"य़",u"ਲ਼",u"ਸ਼",u"ਖ਼",u"ਗ਼",u"ਜ਼",u"ਫ਼",u"ড়",u"ঢ়",u"য়",u"ଡ଼",u"ଢ଼"]

    nuktaDecom = ["\u0915\u093C","\u0916\u093C","\u0917\u093C","\u091C\u093C","\u0921\u093C","\u0922\u093C","\u092B\u093C","\u092F\u093C","\u0A32\u0A3C","\u0A38\u0A3C","\u0A16\u0A3C","\u0A17\u0A3C","\u0A1C\u0A3C","\u0A2B\u0A3C","\u09A1\u09BC","\u09A2\u09BC","\u09AF\u09BC","\u0B21\u0B3C","\u0B22\u0B3C"]
    nuktaPrecom = ["\u0958","\u0959","\u095A","\u095B","\u095C","\u095D","\u095E","\u095F","\u0A33","\u0A36","\u0A59","\u0A5A","\u0A5B","\u0A5E","\u09DC","\u09DD","\u09DF","\u0B5C","\u0B5D"]

    if Source not in ["Grantha","TamilGrantha"]:
        for x,y in zip(nuktaDecom,nuktaPrecom):
            Strng = Strng.replace(x,y)

    if Source in ['IAST', 'ISO', 'ISOPali', 'Titus', 'IASTPali']:
        Strng = Strng.replace("ü", "uʼ").replace("ǖ", "ūʼ").replace( 'ö', 'aʼ',).replace('ȫ', 'āʼ')

    if Source in ['IAST', 'ISO', 'ISOPali', 'IASTPali', 'Titus'] or 'RomanLoC' in Source:
        Strng = unicodedata.normalize('NFC', Strng)
        Strng = Strng.replace("ẗ", "ẗ").replace("ÿ", "ÿ").replace("ḧ", "ḧ")

        if 'RomanLoC' in Source:
            Strng = Strng.replace('\u02D9', '\u02DA')

    if Source == 'Arab-Ur' or Source == 'Arab-Pa':
        Strng = Strng.replace('ك', 'ک')
        Strng = Strng.replace('ي', 'ی')

    if Source == "Hebr":
        vowels = ['ְ','ֱ','ֲ','ֳ','ִ','ֵ','ֶ','ַ','ָ','ֹ','ֺ','ֻ','ׇ']
        vowelsR = '(' + '|'.join(vowels + ['וֹ', 'וּ']) + ')'

        # Swap the order of diacritics
        Strng = re.sub(vowelsR + '([ּׁׂ])', r'\2\1', Strng)
        Strng = Strng.replace('\u05BC\u05B0\u05C1', '\u05C1\u05BC\u05B0')

    # Sindhi C+Anudatta <- Sindhi Underscore characters
    # For easy Transliteration

    # Malayalam Chillu Characters

    chilluZwj=["ണ്‍","ന്‍","ര്‍","ല്‍","ള്‍","ക്‍"]
    chilluAtom=["ൺ","ൻ","ർ","ൽ","ൾ","ൿ"]

    for x,y in zip(chilluZwj,chilluAtom):
        Strng = Strng.replace(x,y)

    Strng = Strng.replace('ൌ', 'ൗ')
    Strng = Strng.replace('ൟ', 'ഈ')
    Strng = Strng.replace('ൎ', 'ര്')

    # Malayalam chill + virama to just letter + virama

    Strng = Strng.replace('ൻ്റ', 'ന്റ')

    # Telugu/Kananda Nakaarapollu

    Strng = Strng.replace('ೝ', 'ನ್') # Kannada
    Strng = Strng.replace('ౝ', 'న్') # Telugu

    # Bengali Khanda Ta

    # Strng = Strng.replace("ৎ","ত্‍")

    # Tamil

    tamAlt = ["ஸ்ரீ","க்‌ஷ","ர‌ி","ர‌ீ"]
    tamNorm = ["ஶ்ரீ","க்ஷ","ரி","ரீ"]

    # Burmese

    Strng = Strng.replace("ဿ", "သ္သ")

    # Tamil

    Strng.replace("ஸ²", "ஶ")

    ## The following is a reversal of custom options check !!!!

    subNum = ["¹","₁","₂","₃","₄"]
    supNum = ["","","²","³","⁴"]

    for x,y in zip(tamAlt+subNum,tamNorm+supNum):
        Strng = Strng.replace(x,y)

    # Tibetan Vowels

    oldVow = ["ྲྀ","ཷ","ླྀ","ཹ","ཱི","ཱུ","ཀྵ","ྐྵ"]
    newVow= ["ྲྀ","ྲཱྀ","ླྀ","ླཱྀ","ཱི","ཱུ","ཀྵ","ྐྵ"]

    for x,y in zip(oldVow,newVow):
        Strng = Strng.replace(x,y)

    # Decomposed to Precomposed Roman Characters

    latinDecom= ["ā","ī","ū","ē","ō","ṃ","ṁ","ḥ","ś","ṣ","ṇ","ṛ","ṝ","ḷ","ḹ","ḻ","ṉ","ṟ"]
    latinPrecom = ["ā","ī","ū","ē","ō","ṃ","ṁ","ḥ","ś","ṣ","ṇ","ṛ","ṝ","ḷ","ḹ","ḻ","ṉ","ṟ"]

    for x,y in zip(latinDecom,latinPrecom):
        Strng = Strng.replace(x,y)

    # Thai Decomposed AM to Precomposed AM

    Strng = Strng.replace("ํา","ำ")

    # For Lao

    Strng  = Strng.replace("ໍາ", 'ຳ')

    ## Two Single Danda to Double Danda

    Strng = Strng.replace("।।","॥")

    ## Decomposed Limbu to precomposed

    Strng = Strng.replace("ᤠ᤺ᤣ", "ᤠᤣ᤺")
    Strng = Strng.replace("᤺ᤣ", "ᤣ᤺")
    Strng = Strng.replace("ᤠᤣ", "ᤥ")

    ## Replace Thai ฎ

    Strng = Strng.replace("ฎ", "ฏ")

    ## Replace Grantha old au with new au

    Strng = Strng.replace('𑍌', '𑍗')

    ## Replace Tibetan Chandra with Nada to the normal one

    Strng = Strng.replace('\u0F82', '\u0F83')

    ## Replace Candra A iwht regualr AE

    Strng = Strng.replace('ॲ', 'ऍ')

    #Strng = Strng.replace('', "ຣ\uE00A")

    ## Normalization for Bengali, Tamil, Malayalam and Grantha

    # Bengali o/au

    Strng = Strng.replace('ো', 'ো')
    Strng = Strng.replace('াে', 'ো')

    Strng = Strng.replace('ৌ', 'ৌ')
    Strng = Strng.replace('ৗে', 'ৌ')

    # Tamil o, O, au

    Strng = Strng.replace('ொ', 'ொ')
    Strng = Strng.replace('ாெ', 'ொ')

    Strng = Strng.replace('ோ', 'ோ')
    Strng = Strng.replace('ாே', 'ோ')

    Strng = Strng.replace('ௌ', 'ௌ')
    Strng = Strng.replace('ௗெ', 'ௌ')

    # Malayalam

    Strng = Strng.replace('ൊ', 'ൊ')
    Strng = Strng.replace('ാെ', 'ൊ')

    Strng = Strng.replace('ോ', 'ോ')
    Strng = Strng.replace('ാേ', 'ോ')

    # Grantha

    Strng = Strng.replace('𑍋', '𑍋')
    Strng = Strng.replace('𑌾𑍇', '𑍋')

    return Strng

# Remove ZWJ/ZWNJ characters
def removeZW(Strng):
    Strng = Strng.replace("\u200C").replace("\u200D")

    return Strng

def PhagsPaArrange(Strng,Source):
    if Source in GM.IndicScripts:
        ListC = "|".join(sorted(GM.CrunchSymbols(GM.Consonants, Source),key=len,reverse=True)) #Do this for all
        ListV = "|".join(sorted(GM.CrunchSymbols(GM.Vowels, Source),key=len,reverse=True))
        ListVS = "|".join(sorted(GM.CrunchSymbols(GM.VowelSignsNV, Source),key=len,reverse=True))
        ListCS = "|".join(sorted(GM.CrunchSymbols(GM.CombiningSigns, Source),key=len,reverse=True))

        vir = GM.CrunchSymbols(GM.VowelSigns,Source)[0]

        yrv = "|".join([GM.CrunchSymbols(GM.Consonants, Source)[i] for i in [25,26,28]])

        Strng = re.sub("("+ListC+")"+"("+vir+")"+"("+yrv+")"+"("+"("+ListVS+")?"+"("+ListCS+")?"+")",r' \1\2\3\4',Strng)
        Strng = re.sub("("+ListC+ListV+")"+"("+"("+ListVS+")?"+"("+ListCS+")?"+")"+"("+ListC+")"+"("+vir+")"+"(?!\s)",r"\1\2\5\6 ",Strng)
        Strng = re.sub("("+ListC+ListV+")"+"("+"("+ListVS+")?"+"("+ListCS+")?"+")"+"("+ListC+")"+'(?!'+vir+')',r"\1\2 \5",Strng)
        Strng = re.sub("("+ListC+ListV+")"+"("+"("+ListVS+")?"+"("+ListCS+")?"+")"+"("+ListC+")"+'(?!'+vir+')',r"\1\2 \5",Strng)

    elif Source in GM.LatinScripts:
        pass

    return Strng

def TamilTranscribeCommon(Strng, c = 31):
    script = "Tamil"

    ListC = GM.CrunchList('ConsonantMap',script)
    ListSC = GM.CrunchList('SouthConsonantMap',script)
    vir = GM.CrunchSymbols(GM.VowelSigns, script)[0]

    ConUnVoiced = [ListC[x] for x in [0,5,10,15,20]]
    ConVoicedJ =  [ListC[x] for x in [2,7,12,17,22]]
    ConVoicedS =  [ListC[x] for x in [2,31,12,17,22]]

    ConNasalsAll = '|'.join([ListC[x] for x in [4, 9, 14, 19, 24]])
    conNasalCa = '|'.join([ListC[x] for x in [9]])
    ConNasalsGroup = [ConNasalsAll, conNasalCa, ConNasalsAll, ConNasalsAll, ConNasalsAll]

    ConMedials = '|'.join(ListC[25:28]+ListSC[0:2]+ListSC[3:4])
    Vowels = '|'.join(GM.CrunchSymbols(GM.Vowels+GM.VowelSignsNV, script))
    Aytham = GM.CrunchList('Aytham',script)[0]
    Consonants = '|'.join(GM.CrunchSymbols(GM.Consonants,script))
    NRA =ListSC[3] + vir + ListSC[2]
    NDRA = ListC[14] + vir + ListC[12] + vir + ListC[26]

    ### Check Siva Siva Mails
    ### Do something about Eyelash ra in Transliterated text

    for i in range(len(ConUnVoiced)):
        pass
        #Strng = re.sub('('+Vowels+Consonants+')'+ConUnVoiced[i]+'('+Vowels+Consonants+')',r'\1'+ConVoicedS[i]+r'\2',Strng)
        Strng = re.sub('('+Vowels+'|'+Consonants+'|'+Aytham+')'+ConUnVoiced[i]+'(?!'+vir+')',r'\1'+ConVoicedS[i],Strng)
        Strng = re.sub('([³])'+ConUnVoiced[i]+'(?!'+vir+')',r'\1'+ConVoicedS[i],Strng)
        Strng = re.sub('³+', '³', Strng)

        # Nasals + Unvoiced -> voiced
        # Rule applied even if spaced -> ulakaJ cuRRu -> ulakaJ juRRu; cenJu -> senju
        # Strng = re.sub('('+ConNasalsSpace+')'+'('+vir+')'+'( ?)'+ConUnVoiced[i],r'\1\2\3'+ConVoicedJ[i],Strng)

        # Only without space : vampu -> vambu;  varim panam -> varim panam
        Strng = re.sub('('+ConNasalsGroup[i]+')'+'('+vir+')'+ConUnVoiced[i],r'\1\2'+ConVoicedJ[i],Strng)

        # Medials + Unvoiced -> Voiced
        Strng = re.sub('('+ConMedials+')'+'('+vir+')'+ConUnVoiced[i]+'(?!'+vir+')',r'\1\2'+ConVoicedS[i],Strng)

    Strng = Strng.replace(NRA,NDRA)

    # Intervocalic S
    Strng = re.sub('(?<!'+'('+ListC[5]+'|'+ListSC[2] +'|' + 'ட' + ')'+vir+')'+ListC[5]+'(?!'+vir+')',ListC[c],Strng)

    import string

    punct = "|".join(['\\'+x for x in list(string.punctuation.replace(".","").replace("?",""))])+"|\s"

    # CA + Spac | Punct + SA -> CCA
    Strng = re.sub('('+ListC[5]+vir+')'+'(('+punct+')+)'+'('+ListC[c]+')',r'\1\2'+ListC[5],Strng)

    # JA + Spac | Punct + SA -> JCA
    Strng = re.sub('('+ListC[9]+vir+')'+'(('+punct+')+)'+'('+ListC[c]+')',r'\1\2'+ListC[7],Strng)

    # GA + Spac | Punct + KA ->
    Strng = re.sub('('+ListC[4]+vir+')'+'(('+punct+')+)'+'('+ListC[0]+')',r'\1\2'+ListC[2],Strng)

    # NA + Spac | Punct + TTA ->
    Strng = re.sub('('+ListC[14]+vir+')'+'(('+punct+')+)'+'('+ListC[10]+')',r'\1\2'+ListC[12],Strng)

    # NA + Spac | Punct + TA ->
    Strng = re.sub('('+ListC[19]+vir+')'+'(('+punct+')+)'+'('+ListC[15]+')',r'\1\2'+ListC[17],Strng)

    # K + k -> hh
    Strng = Strng.replace(Tamil.Aytham[0]+ListC[0],ListC[32]+vir+ListC[32])

    # K -> H
    Strng = Strng.replace(Tamil.Aytham[0],ListC[32]+vir)

    # RR + RRA -> TT+RA
    Strng = re.sub(ListSC[2]+vir+ListSC[2],ListC[10]+vir+ListC[26],Strng)

    # RR | TT + /s + SA -> RR + /s + CA
    Strng = re.sub("("+'['+ListC[10]+ListSC[2]+']'+vir+')'+'(\s)'+'('+ListC[c]+')',r'\1\2'+ListC[5],Strng)

    ## NNN to N, RR to R

    Strng = Strng.replace(ListSC[3],ListC[19])
    #Strng = Strng.replace(ListSC[2],ListC[26])

    return Strng

def TamilTranscribe(Strng):
    Strng = TamilTranscribeCommon(Strng)

    return Strng

def TamilTranscribeDialect(Strng):
    Strng = TamilTranscribeCommon(Strng, c=29)

    return Strng

def IPAIndic(Strng):
    Strng = Strng.replace("ʊ","u")
    Strng = Strng.replace("ɛ","e")

    return



from aksharamukha import transliterate

import unittest

test_strings_rev = {
  'Assamese': [('কৎ', 'kaṯ'), ('কৱ কষ', 'kawa kasha'), ('কঽ কে কো কৃ কং ওঁ', 'ka’ ke ko kr̥ kaṃ ōṃ'), ('কঁদ কঁপ কাঁ কাঁশ', 'kan̐da kan̐pa kām̐ kām̐śa'), ('কংদ কংপ', 'kaṃda kaṃpa')],
  'Balinese': [('ᬅᬜ ᬅᬗ ᬅᬯ ᬗ᬴ ᬗ᬴ᬷ ᬧ᬴ᭀ ᬚ᬴ᬾ', 'aña anga awa ‘a ‘ī fo ze'), \
               ('ᬅᭂ ᬅᭃ ᬓᭂ ᬓᭃ ᬋ ᬌ ᬍ ᬎ ᬓᬼ ᬓᬽ', 'ĕ ö kĕ kö rĕ rö lĕ lö klĕ klö'), \
                ('ᬓᬃᬧ ᬓᬧ᭄ᬭ ᬓᬧ᭄ᬬ ᬓᬵᬃᬬ', 'karpa kapra kapya kārya'), ('ᬓᬾ ᬓᭀ', 'ke ko')],
  'Bengali': [('কৎ', 'kaṯ'), ('কষ', 'kasha'), ('কঽ কে কো কৃ কং ওঁ', 'ka’ ke ko kr̥ kaṃ ōṃ'), ('কঁদ কঁপ কাঁ কাঁশ', 'kan̐da kan̐pa kām̐ kām̐śa'), ('কংদ কংপ', 'kaṃda kaṃpa')],
  'Gujarati': [('ષિષી કં કળત્ર કંય કં ઍ કૅ ઑ કૉ', 'shishī kaṃ kaḷatra kaṃya kaṃ ê kê ô kô'), ('કે કો કૃ કં ૐ', 'ke ko kr̥ kaṃ oṃ')],
  'Kannada': [('ಕಳ ಕಱೀ ಕೞ', 'kaḷa kaṟī kal̤a')],
  'Limbu': [('ᤀᤣ ᤀᤧ ᤀᤥ ᤀᤨ ᤁ᤹ᤕᤥ ᤕᤢ᤺ᤔᤠ ᤌᤠ᤺ᤒᤧ ᤁᤩ ᤁᤪ ᤁᤫ ᤁᤰ ᤁᤴ ᤁᤵ', 'e ĕ o ŏ kaḥyo yu’mā thā’bĕ kya kra kva kak kan kap')],
  'Malayalam': [('കു് കറ്റ കള കഴ', 'kȧ kaṯṯa kaḷa kaḻa')],
  'Oriya': [('କେ କୋ ୱ ଷ କଁପ କଁଝ କଁ କଁଯ', 'ke ko wa sha kan̐pa kan̐jha kam̐ kam̐ya'), ('କେ କୋ କୃ କଂ ଓଁ', 'ke ko kr̥ kaṃ oṃ')],
  'Devanagari': [('के को कृ कँय कँद कँड', 'ke ko kr̥ kam̐ya kan̐da kan̐ḍa'), ('के को कृ कं ॐ', 'ke ko kr̥ kaṃ oṃ')],
  'Gurmukhi': [('ਕਲੵ ਸ਼ ਖ਼ ਗ਼ ਵ ਅੱਧਕ ਅੱਦਕ', 'kaḷa sha k͟ha g͟ha wa addhaka addaka'), \
               ('ਅੰਯ ਅੰ ਅੰਨ ਅੰ ਕੀਂ', 'am̆ya am̆ am̆na am̆ kīṃ'), ('ਕੇ ਕੋ ਕ੍ਰੁ ਕੰ ੴ', 'ke ko kru kam̆ oṁ')],
  'Sinhala': [('ඇ ඈ කැ කෑ කළ කං කංශ', 'ă â kă kâ kaḷa kaṃ kaṃśa'), \
              ('කේ කෝ කෘ කං ඕං', 'kē kō kr̥ kaṃ ōṃ')],
  'Telugu': [('కే కో కృ కం కంశ కఁ కఁశ', 'kē kō kr̥ kaṃ kaṃśa kam̐ kam̐śa'), ('అఁగ అఁద అఁడ అఁబ అఁజ', 'an̐ga an̐da an̐ḍa an̐ba an̐ja'), ('కే కో కృ కం ఓం', 'kē kō kr̥ kaṃ ōṃ')],
  'Tibetan': [('ཅ ཆ ཇ ཇྷ ཙ ཚ ཛ ཛྷ ཤ ཉ ནྒ ཝ', 'ca cha ja jha tsa tsha dza dzha sha nya nga wa'), \
              ('ཞ ཟ འ ྅', 'zha za ’a `'), ('ཏྶ ནྱ ནྒ ནྟ ཏྭ ཏྐ', 'tʹsa nʹya nga nta twa tka'), \
                ('ཀེ ཀོ ཀྲྀ ཀཾ ༀ', 'ke ko kr̥ kaṃ oṃ')]

}

test_strings_irrev = {
  'Assamese': [],
  'Balinese': [('ᬓ᬴ ᬳ᬴ᬷ ᬕ᬴᭄ᬯ', 'kha hī ghwa'), ('ᬓᬄ ᬓᬂ ᬓᬁ', 'kah kang kang')],
  'Bengali': [('উদ্বেগ উক্বর উব্ব ভ়ুদ', 'udvega ukvara ubba vuda')],
  'Gujarati': [('કંભિ કંઘ', 'kambhi kaṅgha')],
  'Javanese': [('ꦮ ꦝ ꦄꦝ꧀ꦝ ꦄꦚꦹ ꦡ ꦛ ꦄꦛ꧀ꦛ ꦄꦔ ꦬ', 'wa dha adhdha anyu tha tha aththa anga ra'), ('ꦲ꦳ ꦏ꦳ ꦥ꦳ ꦗ꦳ ꦒ꦳ ꦔ꦳', 'ha kha fa za gha ‘a'), ('ꦏꦴ ꦏꦷ ꦏꦹ ꦏꦼ ꦏꦼꦴ ꦏꦺ ꦏꦻ ꦏꦺꦴ ꦏꦻꦴ ꦏꦸ ꦏꦹ ꦏꦁ ꦏꦀ ꦏꦃ ꦄꦏꦿ  ꦄꦂꦏ ꦄꦏ꧀ꦭ ꦄꦏꦽ ꦄꦏ꧀ꦊ ꦄꦏꦾ', 'ka ki ku kĕ kĕ ke ke ko ko ku ku kang kang kah akra  arka akla akrĕ aklĕ akya')],
  'Kannada': [('ಕಂ ಕಂಯ ಕಂಬ ಕಂಟ', 'kaṃ kaṃya kamba kaṇṭa')],
  'Limbu': [],
  'Malayalam': [('കം കംശ കംബ കംധി', 'kaṃ kaṃśa kamba kandhi')],
  'Oriya': [('କଂପ କଂଯ', 'kampa kaṃya')],
  'Devanagari': [('कं कंय कंप', 'kaṃ kaṃya kampa')],
  'Gurmukhi': [('ਅੰਪ ਕੀਂਪ', 'ampa kīmpa')],
  'Sinhala': [('අඟ අඳ අඬ අඹ අඦ', 'aṅga anda aṇḍa amba añja'), ('අංබ අංග', 'amba aṅga')],
  'Telugu': [('అంబ అంగ', 'amba aṅga')]

}

class TestStringMethods(unittest.TestCase):
  def test_indic_roman_loc(self):
    for script, test_string in test_strings_rev.items():
      for script_str, loc_str in test_string:
        self.assertEqual(transliterate.process(script, 'RomanLoC', script_str), loc_str)

    for script, test_string in test_strings_irrev.items():
      for script_str, loc_str in test_string:
        self.assertEqual(transliterate.process(script, 'RomanLoC', script_str), loc_str)

  def test_roman_loc_indic(self):
    for script, test_string in test_strings_rev.items():
      for script_str, loc_str in test_string:
        self.assertEqual(transliterate.process('RomanLoC', script, loc_str, nativize=False, post_options=['TibetanTsheg']), script_str)

  def test_balinese_simple(self):
    script = "ᬡ ᬙ ᬣ ᬰ ᬱ ᬨ ᬖ ᬪ"
    loc = "na ca ta sa sa pa gha ba"
    self.assertEqual(transliterate.process('Balinese', 'RomanLoC', script, post_options=['BalineseSimplified']), loc)

  def test_javanese_simple(self):
    script = "ꦑ ꦡ ꦯ ꦰꦴ ꦥ ꦘ ꦓ ꦨ"
    loc = "ka ta sa sa pa nya ga ba"
    self.assertEqual(transliterate.process('Javanese', 'RomanLoC', script, post_options=['JavaneseSimplified']), loc)

  def test_hindi_marathi(self):
    script = "ऎ कॆ ऍ कॅ कॉ ऑ क़ ख़ ग़ घ़ श ट़ फ़ ष स़ कळ" #ह़
    loc = "ĕ kĕ ê kê kô ô qa k͟ha g͟ha g̳h̳a śa t̤a fa sha s̤a kaḷa" #h̤a

    self.assertEqual(transliterate.process('Devanagari', 'RomanLoC', script, post_options=['HindiMarathiRomanLoCFix']), loc)

    self.assertEqual(transliterate.process('RomanLoC', 'Devanagari', loc, pre_options=['HindiMarathiRomanLoCFix']), script)

if __name__ == '__main__':
  unittest.main()


# -*- coding: utf-8 -*-

# Script Mapping for Devanagari

VowelMap =  [
            '\uABD1',
            '\uABD1\uABE5',
            '\uABCF',
            '\uABD1\uAAEB',
            '\uABCE',
            '\uABD1\uAAEC',
            '\uABD4\uABE8\u02BD',
            '\uABD4\uAAEC\u02BD',
            '\uABC2\uABE8\u02BD',
            '\uABC2\uAAEC\u02BD',
            '\uABD1\uABE6',
            '\uABD1\uABE9',
            '\uABD1\uABE3',
            '\uABD1\uABE7',
            ]

SouthVowelMap = [
                '\uABD1\uABE6\u02BD',
                '\uABD1\uABE3\u02BD',
                ]

ModernVowelMap = [
                 '\uABD1\uABE6\u02BD',
                 '\uABD1\uABE5\u02BD',
                 ]

SinhalaVowelMap = [
                  '\uABD1\uABE6\u02BD'
                  ]

VowelSignMap =  [
                 '\uABE5',
                 '\uABE4',
                 '\uAAEB',
                 '\uABE8',
                 '\uAAEC',
                 '\uABED\uABD4\uABE8\u02BD',
                 '\uABED\uABD4\uAAEC\u02BD',
                 '\uABED\uABC2\uABE8\u02BD',
                 '\uABED\uABC2\uAAEC\u02BD',
                 '\uABE6',
                 '\uABE9',
                 '\uABE3',
                 '\uABE7',
                 ]

SouthVowelSignMap = [
                     '\uABE6\u02BD',
                     '\uABE3\u02BD',
                     ]

ModernVowelSignMap =[
                     '\uABE6\u02BD',
                     '\uABE5\u02BD',
                     ]

SinhalaVowelSignMap = [
                       '\uABE6\u02BD'
                      ]

AyogavahaMap = [
               '\uABEA\u02BD',
               '\uABEA',
               '\uAAF5'
               ]

ViramaMap =  [
             '\uABED'
             ]

ConsonantMap =  [
                '\uABC0',
                '\uABC8',
                '\uABD2',
                '\uABD8',
                '\uABC9',

                '\uABC6',
                '\uAAE2',
                '\uABD6',
                '\uABD3',
                '\uAAE3',

                '\uAAE4',
                '\uAAE5',
                '\uAAE6',
                '\uAAE7',
                '\uAAE8',

                '\uABC7',
                '\uABCA',
                '\uABD7',
                '\uABD9',
                '\uABC5',

                '\uABC4',
                '\uABD0',
                '\uABD5',
                '\uABDA',
                '\uABC3',

                '\uABCC',
                '\uABD4',
                '\uABC2',
                '\uABCB',

                '\uAAE9',
                '\uAAEA',
                '\uABC1',
                '\uABCD'
                ]

SouthConsonantMap = [
                    '\uABC2\u02BD',
                    '\uABC2\u02BD',
                    '\uABD4\u02BD',
                    '\uABC5\u02BD'
                    ]

NuktaConsonantMap =  [
                     '\uABC0\u02BD',
                     '\uABC8\u02BD',
                     '\uABD2\u02BD',
                     '\uABD6\u02BD',
                     '\uABD7\u02BD',
                     '\uABD9\u02BD',
                     '\uABD0\u02BD',
                     '\uABCC\u02BD'
                     ]

SinhalaConsonantMap =[
                     '\uABEA\uABD2\u02BD',
                     '\uABEA\uABD6\u02BD',
                     '\uABEA\uABD7\u02BD',
                     '\uABEA\uABD7\u02BD',
                     '\uABEA\uABD5\u02BD',
                      ]

NuktaMap = [
           '\u02BD\u02BD\u02BD\u02BD'
           ]

OmMap = [
        '\uABD1\uABE3\uABEA'
        ]

SignMap =[
         '\u0027',
         '\uABEB',
         '\uABEB\uABEB'
         ]

Aytham =[AyogavahaMap[2]+'\u02BC']

NumeralMap = [
             '\uABF0',
             '\uABF1',
             '\uABF2',
             '\uABF3',
             '\uABF4',
             '\uABF5',
             '\uABF6',
             '\uABF7',
             '\uABF8',
             '\uABF9'
             ]

from ... import GeneralMap as GM

import os
GM.add_additional_chars(dict([(charlist, globals()[charlist]) for charlist in GM.CharmapLists]),
                        os.path.splitext(__file__)[0].split(os.path.sep)[-1])



# -*- coding: utf-8 -*-

# Script Mapping for ISO

VowelMap =  [
            'â',
            'ā̂', # A
            'î', #i
            'ī̂', #I
            'û', # u
            'ū̂', # U
            'ru\u02BD',
            'rū\u02BD',
            'lu\u02BD',
            'lū\u02BD',
            'ē̂', #e
            'âŷ', #ai
            'ō̂', #o
            'âŵ' #au
            ]

SouthVowelMap = [
                'ê',
                'ô'
                ]

ModernVowelMap = [
                 'ē̂\u02BD',
                 'ā̂\u02BD'
                 ]

SinhalaVowelMap = [
                  'ê\u02BD'
                  ]

VowelSignMap =  [
            'ā', # A
            'i', #i
            'ī', #I
            'u', # u
            'ū', # U
            '\u033Dru\u02BD',
            '\u033Drū\u02BD',
            '\u033Dru\u02BD',
            '\u033Drū\u02BD',
            'ē', #e
            'aŷ', #ai
            'ō', #o
            'aŵ' #au
]

SouthVowelSignMap = ['e','o']

ModernVowelSignMap = ['e\u02BD','ā\u02BD']

SinhalaVowelSignMap = [
                  'ē\u02BD'
                  ]

AyogavahaMap = [
               'ₘ',
               'ₘ', ## Check nasalization
               'h\u033D\u02BD'
               ]

ViramaMap =  [
             '\u033D'
             ]

ConsonantMap =  [
                'k',
                'kʰ',
                'g',
                'gʰ',
                'n\u02BD',

                'č',
                'čʰ',
                'j',
                'jʰ',
                'ɲ',

                'ʈ',
                'ʈʰ',
                'ɖ',
                'ɖʰ',
                'ɳ',

                't',
                'tʰ',
                'd',
                'dʰ',
                'n',

                'p',
                'pʰ',
                'b',
                'bʰ',
                'm',

                'y',
                'r',
                'l',
                'v',

                'š',
                'ʂ',
                's',
                'h',
                ]

SouthConsonantMap = [
                    'ɭ',
                    'ɭ\u02BD',
                    'r\u02BD',
                    'n\u02BD'
                    ]

NuktaConsonantMap =  [
                     'q',
                     'ḫ',
                     'ġ',
                     'z',
                     'ɽ',
                     'ɽʰ',
                     'f',
                     'y\u02BD'
                     ]

SinhalaConsonantMap =[
                     'n\u033Dg\u02BD',
                     'n\u033Dj\u02BD',
                     'n\u033Dd\u02BD',
                     'n\u033Dd\u02BD',
                     'm\u033Db\u02BD',
                      ]

NuktaMap = [
           'Q'
           ]

OmMap = [
        'ō̂m\u033D\u02BD'
        ]

SignMap =[
         '\u02BD\u02BD\u02BD\u02BD',
         '.',
         '..'
         ]

Aytham =['h\u033D\u02BD']

NumeralMap = [
             '\u0030',
             '\u0031',
             '\u0032',
             '\u0033',
             '\u0034',
             '\u0035',
             '\u0036',
             '\u0037',
             '\u0038',
             '\u0039',
             ]

from ... import GeneralMap as GM

import os
GM.add_additional_chars(dict([(charlist, globals()[charlist]) for charlist in GM.CharmapLists]),
                        os.path.splitext(__file__)[0].split(os.path.sep)[-1])


from aksharamukha import transliterate

import unittest

test_strings_rev = [
  ('ᨠᩢᨠ ᨠᩡ ᨠᩥ ᨠᩦ ᨠᩩ ᨠᩪ ᨠᩮ ᨠᩮᩡ ᨠᩰᩡ ᨠᩰ ᨠ᩠ᨿᨠ ᨠ᩠ᩅᩫᩡ ᨠ᩠ᩅᩫ ᨠ᩠ᩅᨠ', 'kaka kaḥ ki kī ku kū ke keḥ koḥ ko kèka kòḥ kò kòka'),
  ('ᨠᩰᩬᩡ ᨠᩫᨠ ᨠᩳ ᨠᩬᨠ ᨠᩯᩡ ᨠᩯ ᨠᩧ ᨠᩨ ᨠᩮᩨᩬᩡ ᨠᩮᩨᩬ ᨠᩮᩨᨠ ᨠᩱ ᨠᩱ᩠ᨿ ᨠ᩠ᩅᩭ ᨠᩭ ᨠᩴ᩠ᨿ ᨠᩮᩢᩣ', 'kǫḥ kǫḥka kǫ kǫka kæḥ kæ kư kư̄ kœḥ kœ kœka kai kaiy koi kǫi keo kao'),
  ('ᩋ ᩋᩣ ᩍ ᩎ ᩏ ᩐ ᩑ ᩋᩡ ᩒᩡ ᩑᩡ', 'a ā i ī u ū e aḥ oḥ eḥ'),
  ('ᩔ ᨬᩚ ᩈ', 'ssa ñña sa'),
  ('ᩋᩢᨠ ᩋᩡ ᩋ᩠ᨿᨠ ᩋ᩠ᩅᩫᩡ ᩋ᩠ᩅᩫ ᩋ᩠ᩅᨠ', 'aka aḥ èka òḥ ò òka'),
  ('ᩋᩰᩬᩡ ᩋᩫᨠ ᩋᩳ ᩋᩬᨠ ᩋᩯᩡ ᩋᩯ ᩋᩧ ᩋᩨ ᩋᩮᩨᩬᩡ ᩋᩮᩨᩬ ᩋᩮᩨᨠ ᩋᩱ ᩋᩱ᩠ᨿ ᩋ᩠ᩅᩭ ᩋᩭ ᩋᩴ᩠ᨿ ᩋᩮᩢᩣ', 'ǫḥ ǫḥka ǫ ǫka æḥ æ ư ư̄ œḥ œ œka ai aiy oi ǫi eo ao'),
  ('ᩍ᩵ ᩍ᩶ ᩍ᩷ ᩍ᩸ ᩍ᩹', 'i′ i″ iˆ iᴶ iʵ')
]

test_strings_irrev = [
 ('ᨠᩤ', 'kā'),
  ('ᨠᨠ᩠ᨿ ᨠᨠ᩠ᩅ', 'kakya kakva')
]

test_strings_irrev = [
]
class TestStringMethods(unittest.TestCase):
  def test_tham_roman_loc(self):
    for tham, loc in test_strings_rev + test_strings_irrev:
      self.assertEqual(transliterate.process('TaiTham', 'RomanLoC', tham), loc)

  def test_roman_loc_tham(self):
    for tham, loc in test_strings_rev:
      self.assertEqual(transliterate.process('RomanLoC', 'TaiTham', loc), tham)

if __name__ == '__main__':
  unittest.main()


"""A setuptools based setup module.

See:
https://packaging.python.org/en/latest/distributing.html
https://github.com/pypa/sampleproject
"""

# To use a consistent encoding
from os import path

# Always prefer setuptools over distutils
from setuptools import setup, find_packages

here = path.abspath(path.dirname(__file__))

import io

with io.open('README.md', 'r', encoding='utf8') as f:
    long_description = f.read()

setup(
  name='aksharamukha',

  # Versions should comply with PEP440.  For a discussion on single-sourcing
  # the version across setup.py and the project code, see
  # https://packaging.python.org/en/latest/single_source_version.html
  version='2.3',

  description='Provides script conversion (a.k.a transliteration) between various scripts',
  long_description=long_description,
  long_description_content_type='text/markdown',

  # The project's main homepage.
  url='https://github.com/virtualvinodh/aksharamukha-python',

  # Author details
  author='Vinodh Rajan',
  author_email='vinodh@virtualvindh.com',

  # Choose your license
  license='GNU AGPL 3.0',

  # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
  classifiers=[
    # How mature is this project? Common values are
    #   3 - Alpha
    #   4 - Beta
    #   5 - Production/Stable
    'Development Status :: 5 - Production/Stable',

    # Indicate who your project is intended for
    'Intended Audience :: Education',
    'Intended Audience :: Science/Research',
    'Intended Audience :: Developers',

    'Natural Language :: Tamil',
    'Natural Language :: Telugu',
    'Natural Language :: Hindi',
    'Natural Language :: Marathi',

    'Natural Language :: Bengali',
    'Natural Language :: Urdu',
    'Natural Language :: Thai',
    'Natural Language :: Tibetan',
    'Natural Language :: Hebrew',
    'Natural Language :: Arabic',

    'Topic :: Text Processing :: Linguistic',

    # Pick your license as you wish (should match "license" above)
    'License :: OSI Approved :: GNU Affero General Public License v3',

    # Specify the Python versions you support here. In particular, ensure
    # that you indicate whether you support Python 2, Python 3 or both.
    'Programming Language :: Python :: 3.8',
    'Programming Language :: Python :: 3.9',
    'Programming Language :: Python :: 3.10',
  ],

  # What does your project relate to?
  keywords='unicode semitic arabic syriac hebrew indic indian language script brahmi brahmic asian transliteration conversion writing alphabet romanization',

  # You can just specify the packages manually here if your project is
  # simple. Or you can use find_packages().
  packages=find_packages(),

  # Alternatively, if you want to distribute just a my_module.py, uncomment
  # this:
  #   py_modules=["my_module"],

  # List run-time dependencies here.  These will be installed by pip when
  # your project is installed. For an analysis of "install_requires" vs pip's
  # requirements files see:
  # https://packaging.python.org/en/latest/requirements.html
  install_requires=['Requests>=2.20.1', 'pykakasi>=2.0.6', 'pyyaml>=5.4.1', 'langcodes>=3.1.0' , 'language_data', \
    'langcodes>=3.1.0', 'language_data', 'regex>=2021.8.3', 'fonttools[unicode]>=4.27', 'lxml'],

  # List additional groups of dependencies here (e.g. development
  # dependencies). You can install these using the following syntax,
  # for example:
  # $ pip install -e .[dev,test]
  # If there are data files included in your packages that need to be
  # installed, specify them here.  If using Python 2.6 or less, then these
  # have to be included in MANIFEST.in as well.
  package_data={
    'aksharamukha': ['yaml/aksharamukha-scripts.yaml', 'yaml/wikitra2-data.yaml', 'json/gimeltra_data.json'],
  },

  # Although 'package_data' is the preferred approach, in some case you may
  # need to place data files outside of your packages. See:
  # http://docs.python.org/3.4/distutils/setupscript.html#installing-additional-files # noqa
  # In this case, 'data_file' will be installed into '<sys.prefix>/my_data'
  # data_files=[('my_data', ['data/data_file'])],

  # To provide executable scripts, use entry points in preference to the
  # "scripts" keyword. Entry points provide cross-platform support and allow
  # pip to create the appropriate form of executable for the target platform.
  # entry_points={
  #     'console_scripts': [
  #         'sample=sample:main',
  #     ],
  # },
)



from aksharamukha import transliterate

import unittest

test_strings_rev = [
    ('ဢ', '’a'),
    ('ၵႃ ၵိ​ ၵီ​ ၵု ၵူ ၵို ၵိူ ၵေ ၵွ ၵႄ​ ၵွႆ ၵႂ် ၵၢႆ ၵေႃ ၵေႃ် ၵျ ၵြ ၵႂ ၵိူဝ် ၵူဝ် ၵဝ်', \
     'kā ki​ kī​ ku kū kui kūi ke kau kè​ koi kái kāi ko kò kya kra kva kūivʻ kō kavʻ'),
    ('ဢႃ ဢိ​ ဢီ​ ဢု ဢူ ဢို ဢိူ ဢေ ဢွ ဢႄ​ ဢွႆ ဢႂ် ဢၢႆ ဢေႃ ဢေႃ် ဢိူဝ် ဢူဝ် ဢဝ်', \
     '’ā ’i​ ’ī​ ’u ’ū ’ui ’ūi ’e ’au ’è​ ’oi ’ái ’āi ’o ’ò ’ūivʻ ’ō ’avʻ'),
    ('တြႃးတေႃ်ၸဝ်သုၵ်သီလၶၼ်း​','trāʺtòcavʻsukʻsīlakhanʻʺ​'),
    ('မွင်းဝၼ်းၶိုၼ်းႁၼ်', 'mauṅʻʺvanʻʺkhuinʻʺhanʻ'),
    ('ဢမ်ႇၶႂ်ႈၼွၼ်းၽၼ်', '’am̢ʻkhái̐naunʻʺphanʻ'),
    ('ၸိူင်းပွတ်းၽွမ်ႉႁူမ်ႈ', 'cūiṅʻʺpautʻʺphaumʻʹhūm̐ʻ'),
    ('မၢႆၼိုင်ႈ', 'māinuiṅ̐ʻ'),
    ('ၼင်းၸံးပူးဝွၵ်း', 'naṅʻʺcaṃʺpūʺvaukʻʺ'),
    ('ပရိတ်တပႃႇလိလႄႈၶေႃႈပွင်ႇၽၢႆႉတႆး', 'paritʻtapā̢lilè̐kho̐pauṅ̢ʻphāiʹtaiʺ'),
    ('ၼံႉၵႂံးၸဝ်သင်လုင်ဝၼ်ဢႆ', 'naṃʹkvaṃʺcavʻsaṅʻluṅʻvanʻ’ai'),
    ('၊ ။', ', .')
]

test_strings_rev_openclosed = [
    ('ၵေးၵဵၸ်ၷႄႉၶႅၺ်ႊပႃႊပၢထ်း','keʺkecʻgèʹkhèñʻ˝pā˝pāthʻʺ'),
    ('မဵတ်ႉတႃႇၸႅတ်ႈ', 'metʻʹtā̢cèt̐ʻ'),
    ('လေႃးၵီႇၵတ်းယဵၼ်', 'loʺkī̢katʻʺyenʻ'),
    ('ယိင်းမူၺ်သႅင်လိူၼ်', 'yiṅʻʺmūñʻsèṅʻlūinʻ'),
    ('ပိုၼ်းၽြႃးတႃႇလုၵ်ႈႁဵၼ်းပၢၼ်မႂ်ႇ', 'puinʻʺphrāʺtā̢luk̐ʻhenʻʺpānʻmái̢'),
    ('ၸၢတ်ႈၸဝ်ႈသုဝၼ်ႇၼသၢမ်ႇ', 'cāt̐ʻcav̐ʻsuvan̢ʻnasām̢ʻ'),
    ('လႅင်းမုၼ်းမႂ်ႇ', 'lèṅʻʺmunʻʺmái̢'),
    ('ၵေႇၵဵၵ်းၻႄႉၻႅသ်ႊတႃႈႀၢရ်း', 'ke̢kekʻʺdèʹdèsʻ˝tā̐xārʻʺ')
]

test_strings_irrev = [
    ('ပပ်ႉၽိုၵ်းလႃတ်ႈၵႂႃမ်းႁႃႈၽႃႇသႃႇ', 'papʻʹphuikʻʺlāt̐ʻkvāmʻʺhā̐phā̢sā̢'),
    ('တႃႇႁဵတ်းၵႃၼ်ၶိုၼ်ႈယႂ်ႇ', 'tā̢hetʻʺkānʻkhuin̐ʻyái̢'),
    ('ၶႅပ်းမႃင်ၸဝ်ႈꩪမ်ႇမၻႃႉၼ', 'khèpʻʺmāṅʻcav̐ʻdham̢ʻmadāʹna'),
    ('ၿႆၢ ၵၢ ၵႃႆ ၵေေ ၵေါ','bāi kā kāi kè ko')
]

class TestStringMethods(unittest.TestCase):
    def test_shan_roman_loc(self):
        for shan, loc in test_strings_rev + test_strings_rev_openclosed:
            self.assertEqual(transliterate.process('Shan', 'RomanLoC', shan), loc)

    def test_roman_loc_shan(self):
        for shan, loc in test_strings_rev + test_strings_rev_openclosed:
            self.assertEqual(transliterate.process('RomanLoC', 'Shan', loc), shan)

    def test_shan_roman_loc_irrev(self):
        for shan, loc in test_strings_irrev:
            self.assertEqual(transliterate.process('Shan', 'RomanLoC', shan), loc)

if __name__ == '__main__':
    unittest.main()


# -*- coding: utf-8 -*-

from aksharamukha import Convert as cn
import xml.etree.ElementTree as ET
from aksharamukha import GeneralMap as GM
import re

def GetSyllables(Strng,Script):

    if Script not in GM.IndicScripts:

        ListV = '|'.join(sorted(GM.CrunchSymbols(GM.Vowels,Script),key=len,reverse=True))
        ListC = '|'.join(sorted(GM.CrunchSymbols(GM.Consonants,Script),key=len,reverse=True))
        ListA = '|'.join(sorted(GM.CrunchSymbols(GM.CombiningSigns,Script),key=len,reverse=True))

        Sylbl = re.compile('(('+ListC+')*'+'(('+ListV+')('+ListA+')?)?)')

        ListSyl = [Mtch[0] for Mtch in Sylbl.findall(Strng) if Mtch[0] != '']

    return ListSyl

def GetSidRanCh(Char,Script,reverse=False):

    if Script is 'Siddham':
        tree = ET.parse('C:\Personal Projects\\aksharamukha\ScriptMap\EastIndic\siddham.xml')
        tbl = 'tbl_siddham'
        defChar = '袎'
    elif Script is 'Ranjana':
        tree = ET.parse('C:\Personal Projects\\aksharamukha\ScriptMap\EastIndic\\ranjana.xml')
        tbl = 'tbl_lantsa'
        defChar = '跬'

    root = tree.getroot()
    fnd = False

    for mapng in root.findall(tbl):
        for ch in mapng:
            if ch.text == Char:
                fndmap = mapng
                fnd = True
        if fnd:
            break

    if fnd:
        if not reverse:
            return fndmap.find('chars').text
        else:
            return fndmap.find('ime').text
    else:
        return defChar

def SiddhRanjConv(Strng,Script,reverse=False):
    if not reverse:
        Sylbl = sorted(GetSyllables(Strng,'HK'),key=len,reverse=True)

        SylblCon = [GetSidRanCh(x,Script) for x in Sylbl if x != '']

        for x,y in zip(Sylbl,SylblCon):
            Strng = Strng.replace(x,y)

        Strng = Strng.replace('_','')

    else:
        CharList = list(set(Strng.replace(" ","")))
        if Script == "Siddham":
            Strng = Strng.replace("袎","(.)")
        else:
            Strng = Strng.replace("跬","(.)")
        for char in CharList:
            Strng = Strng.replace(char,GetSidRanCh(char,Script,reverse=True))

    return Strng

if __name__ == "__main__":
    print(cn.convertScript("namo", "HK", "Siddham"))
    print(cn.convertScript("巧伕", "Siddham", "Telugu"))
    print(cn.convertScript("巧伕", "Siddham", "Ranjana"))
    print(cn.convertScript("呣呺","Ranjana","Kannada"))



# -*- coding: utf-8 -*-

from . import GeneralMap as GM, PreProcess as PrP, ConvertFix as CF
from . import PostProcess as PP
import aksharamukha.ScriptMap.EastIndic.SiddhamRanjana as SR
import aksharamukha.FallBack as FB
import string
import re
from functools import cmp_to_key
import json
from . import gimeltra

### Mapping : https://viss.wordpress.com/2015/05/17/how-to-transcribe-pa%E1%B8%B7i-in-lanna-and-burmese/ ###
### Assmae kva -> becomes kba ## Check Assamese Wikipedia
### ITRANS features like comments implement
### Check Aksharamukha Todo list... and cross check and implement that
### Add HK-Itrans Option of Vishvas Vasuki
### Remove Diacritics
### Add space before punctiations and then remove it

# Sort Functions

def TamilSort(x,y):
    if('\u0B83' in x[0] and len(x[0]) != 1):
        return -1
    elif(x[0] < y[0]):
        return 1
    else:
        return 0

def lenSort(x,y):
    if(len(x[0]) > len(y[0])):
        return -1
    else:
        return 0

def convertInter(Strng,Source):
    ScriptAll = GM.Vowels+GM.Consonants+GM.CombiningSigns+GM.Numerals+GM.Signs+GM.Aytham
    SourceScript = GM.CrunchSymbols(ScriptAll,Source)
    TargetScript = GM.CrunchSymbols(ScriptAll,GM.Inter)
    ScriptMapAll = sorted(zip(SourceScript,TargetScript),key=cmp_to_key(lenSort))

    for x,y in ScriptMapAll:
        Strng = Strng.replace(x,y)

    #print(Strng)

    return Strng

# Conversion Module
def convertScript(Strng,Source,Target):
    #print("I am here")

    charPairs=[];
    Schwa = '\uF000'
    DepV = '\u1E7F'


    ## Itrans Comments ##

    if Source in GM.LatinScripts and Target in GM.IndicScripts:
        try:
            Strng = getattr(CF,"Fix"+Source)(Strng,reverse=True)
        except AttributeError:
            pass
            #print #"Fix"+Target+" doesn't exist - Reverse"

       #short u
        if Source in ['IAST', 'ISO', 'ISOPali', 'Titus']:
            Strng = Strng.replace('ŭ', 'u\u00D7') # special explicit Virama

        ## Joiners {} : ZWNJ : () : ZWJ
        Strng = Strng.replace("{}", "\u200C")
        Strng = Strng.replace("()", "\u200D")

        Strng = CF.VedicSvarasLatinIndic(Strng, Source)

        punc =  '(' + '|'.join(["\u005C"+x for x in list(string.punctuation)]+ ['\s']
                    + [x.replace('.', '\.') for x in GM.CrunchSymbols(GM.Signs,Source)[1:3]]) + ')'

        sOm, tOm = GM.CrunchList('OmMap', Source)[0],GM.CrunchList('OmMap', Target)[0]

        Strng = re.sub(punc + sOm + punc, r'\1' + tOm + r'\2', Strng)
        Strng = re.sub('^' + sOm + punc, tOm + r'\1', Strng)
        Strng = re.sub(punc + sOm + '$', r'\1' + tOm, Strng)
        Strng = re.sub('^' + sOm + '$', tOm, Strng)

        punc = '(\s)'

        Strng = re.sub(punc + sOm + punc, r'\1' + tOm + r'\2', Strng)
        Strng = re.sub('^' + sOm + punc, tOm + r'\1', Strng)
        Strng = re.sub(punc + sOm + '$', r'\1' + tOm, Strng)
        Strng = re.sub('^' + sOm + '$', tOm, Strng)

        SourceOld = Source

        Strng = convertInter(Strng,Source)
        Source = GM.Inter
        Strng = PrP.RomanPreFix(Strng,Source)

        #print(Strng)
        ## HK l_R l_RR

        Strng = Strng.replace("ṿ×_","ṿ")
        Strng = Strng.replace("ṿ×_","ṿ")

        ha = GM.CrunchSymbols(GM.Consonants,Source)[32]
        charPairs=[]

        for charList in GM.ScriptAll:
            # Crunch all related characters into a list
            TargetScript = GM.CrunchSymbols(GM.retCharList(charList),Target)
            if charList == 'VowelSigns':
                # Add DepVSign to all VowelSigns to differentiate from Independent Vowels
                SourceScript = [DepV+x for x in GM.CrunchSymbols(GM.VowelSigns,Source)]
            else:
                SourceScript = GM.CrunchSymbols(GM.retCharList(charList),Source)
            # Create a Tuple for the conversion pair
            ScriptMap = list(zip(SourceScript,TargetScript))
            # Sort the mapping in descending order. Longer Characters are to be replaced first. ऍˇ > ऍ
            ScriptMap.sort(reverse=True);
            charPairs= charPairs + ScriptMap

        charPairs = sorted(charPairs,key=cmp_to_key(lenSort))

        # Perform replacement sequentially for each character group
        for x,y in charPairs:
            #print(x, '-->', y)
            Strng = Strng.replace(x,y)
            #print(Strng)

        ## a_i => a<dev>i<dev> ; a_u = a<dev>u<dev>
        Strng=Strng.replace('_' + GM.CrunchSymbols(GM.Vowels,Target)[2],  GM.CrunchSymbols(GM.Vowels,Target)[2])
        Strng=Strng.replace('_' + GM.CrunchSymbols(GM.Vowels,Target)[4],  GM.CrunchSymbols(GM.Vowels,Target)[4])

        ## Joiners Vir + ZWJ
        vir = GM.CrunchList('ViramaMap', Target)[0]
        Strng = Strng.replace(vir + "[]", "\u200D" + vir)

        if Source in ['Inter']:
            Strng = Strng.replace('\u00D7', vir) # special explicit Virama

        #print Strng

        #print(Strng)

        # Apply Fixes on the Output based on the Script
        Strng = CF.FixIndicOutput(Strng, Source, Target)

        #print(Strng)

    elif Source in GM.LatinScripts and Target in GM.LatinScripts:
        try:
            Strng = getattr(CF,"Fix"+Source)(Strng,reverse=True)
        except AttributeError:
            pass
            #print #"Fix"+Target+" doesn't exist - Reverse"

        ScriptAll = GM.Vowels+GM.Consonants+GM.CombiningSigns+GM.Numerals+GM.Signs+GM.Aytham

        Strng = convertInter(Strng,Source)

        SourceScript = GM.CrunchSymbols(ScriptAll,GM.Inter)
        TargetScript = GM.CrunchSymbols(ScriptAll, Target)
        ScriptMapAll = list(zip(SourceScript,TargetScript))

        for x,y in ScriptMapAll:
            Strng = Strng.replace(x,y)

        Strng = CF.PostFixRomanOutput(Strng,Source,Target)

    elif Source in GM.IndicScripts and Target in GM.IndicScripts:
        Strng = PrP.RemoveJoiners(Strng)

        Strng = CF.ShiftDiacritics(Strng,Source,reverse=True)
        try:
            Strng = getattr(CF,"Fix"+Source)(Strng,reverse=True)
        except AttributeError:
            pass
            #print #"Fix"+Target+" doesn't exist - Reverse"

        punc =  '(' + '|'.join(["\u005C"+x for x in list(string.punctuation)]+ ['\s']
                    + [x.replace('.', '\.') for x in GM.CrunchSymbols(GM.Signs,Source)[1:3]]) + ')'

        sOm, tOm = GM.CrunchList('OmMap', Source)[0],GM.CrunchList('OmMap', Target)[0]

        if len(sOm) != 1:
            Strng = re.sub(punc + sOm + punc, r'\1' + tOm + r'\2', Strng)
            Strng = re.sub('^' + sOm + punc, tOm + r'\1', Strng)
            Strng = re.sub(punc + sOm + '$', r'\1' + tOm, Strng)
            Strng = re.sub('^' + sOm + '$', tOm, Strng)

        if len(sOm) == 1:
            Strng = Strng.replace(sOm, tOm)

        # Iterate for each character group
        for charList in GM.ScriptAll:
            # Crunch all related characters into a list
            SourceScript = GM.CrunchSymbols(GM.retCharList(charList),Source)
            TargetScript = GM.CrunchSymbols(GM.retCharList(charList),Target)
            # Create a Tuple for the conversion pair
            ScriptMap = list(zip(SourceScript,TargetScript))
            # Sort the mapping in descending order. why ?
            ScriptMap.sort(reverse=True)
            charPairs= charPairs + ScriptMap

        #Sort based on Length - Longest first
        charPairs = sorted(charPairs,key=cmp_to_key(lenSort))

        #print(Strng)

        # Perform replacement sequentially for each character group
        for x,y in charPairs:
            # print(x,y)
            Strng = Strng.replace(x,y)
            # print(Strng)

        # print(Strng)

        #Strng = Strng.replace(GM.CrunchList('OmMap', Source)[0],GM.CrunchList('OmMap', Target)[0])
        # Apply Fixes on the Output based on the Script

        Strng = CF.FixIndicOutput(Strng, Source, Target)

    elif Source in GM.IndicScripts and Target in GM.LatinScripts:
        #print(Strng)
        Strng = PrP.RemoveJoiners(Strng)
        Strng = CF.ShiftDiacritics(Strng, Source, reverse=True)
        #print(Strng)
        try:
            Strng = getattr(CF,"Fix"+Source)(Strng,reverse=True)
        except AttributeError:
            pass
            #print #"Fix"+Target+" doesn't exist - Reverse"
        #print(Strng)
        sOm, tOm = GM.CrunchList('OmMap', Source)[0],GM.CrunchList('OmMap', Target)[0]

        Strng = Strng.replace(sOm, tOm)

        # Iterate for each character group
        for charList in GM.ScriptAll:
            # Crunch all related characters into a list
            SourceScript = GM.CrunchSymbols(GM.retCharList(charList),Source)
            if charList == 'Consonants':
                # Add Schwa to all Roman consonants. Basically, the Roman script is "Indianized"
                TargetScript = [x+Schwa for x in GM.CrunchSymbols(GM.Consonants,Target)]
            elif charList == 'Vowels':
                # Add DepVSign to all Independent vowel to differentiate from vowel sign.
                TargetScript = [DepV+x for x in GM.CrunchSymbols(GM.Vowels,Target)]
            else:
                TargetScript = GM.CrunchSymbols(GM.retCharList(charList),Target)
            # Create a Tuple for the conversion pair
            ScriptMap = list(zip(SourceScript,TargetScript))
            # Sort the mapping in descending order. Longer Characters are to be replaced first. ऍˇ > ऍ
            ScriptMap.sort(reverse=True);
            charPairs= charPairs + ScriptMap

        charPairs = sorted(charPairs,key=cmp_to_key(lenSort))

        # special case just for roman semitic
        if Source == 'RomanSemitic':
            unasp = ['k', 'g', 'c', 'j', 't', 'd', 'p', 'b', 'ɽ', 'ʈ', 'ɖ', 'r']
            charPairsH = [(x, y) for x, y in charPairs if 'ʰ' in x]
            charPairsNotH = [(x, y) for x, y in charPairs if 'ʰ' not in x]
            charPairs = charPairsNotH + charPairsH
            for x, y in charPairs:
                if x in unasp:
                    Strng = re.sub(x + '(?!(ʰ|\u0324))', y, Strng)
                else:
                    Strng = Strng.replace(x,y)
        # for all other transformations
        else:
            # Perform replacement sequentially for each character group
            #print(Strng)
            for x,y in charPairs:
                Strng = Strng.replace(x,y)


        # Remove all intermediate characters and fix Output
        Strng = CF.FixRomanOutput(Strng,Target)

        Strng = CF.VedicSvarsIndicLatin(Strng)

        Strng = CF.PostFixRomanOutput(Strng,Source,Target)

        # Convert Syllabic lR -> l_R Important !!!

    elif Source in GM.SemiticScripts and Target in GM.SemiticScripts:
        try:
            Strng = getattr(CF,"Fix"+Source.replace('-', '_'))(Strng,Source,reverse=True)
        except AttributeError:
            pass

        tr = gimeltra.Transliterator()

        if Source == 'Ugar':
            Strng = Strng.replace('𐎟', ' ') # reverse ugaritic word separator

        #print(Strng)
        Strng = tr.tr(Strng, sc=Source, to_sc=Target)
        #print(Strng)

        # Apply Fixes on the Output based on the Script
        Strng = CF.FixSemiticOutput(Strng, Source, Target)

    elif Source in (GM.IndicScripts + GM.LatinScripts) and Target in GM.SemiticScripts:
        tr = gimeltra.Transliterator()

        Strng = convertScript(Strng, Source, "RomanSemitic")

        # print("The string is", Strng)

        #print(Strng)

        Strng = Strng.replace('QQ', '').replace('mQ', '') ## avoiding Q, mQ for Urdu to Semitic : Check why

        # remove gemination for every script except......
        ## Add all scripts with gemination sign here or vowel killer here
        ## Syriac doesn't show gemination hence not added below
        if 'Arab' not in Target and 'Hebr' not in Target and 'Latn' not in Target:
            Strng = re.sub('(.)' + '\u033D' + r'\1', r'\1', Strng)

        Strng = PP.FixSemiticRoman(Strng, Target)

        # for vocalized scripts
        # insert explicit 'a'
        # RomanSemitic is mapped as an Indic script : So no explicit a is added
        # hence /a/ is added explicitly
        if 'Arab' in Target or Target in ['Hebr', 'Syre', 'Syrj', 'Syrn', 'Thaa']:
            Strng = PP.insertARomanSemitic(Strng)

        # remove Virama
        # Strng = Strng.replace('×', '')

        Strng = tr.tr(Strng, sc='Latn', to_sc=Target)

        # Apply Fixes on the Output based on the Script
        Strng = CF.FixSemiticOutput(Strng, Source, Target)

    elif Source in GM.SemiticScripts and Target in (GM.IndicScripts + GM.LatinScripts):
        try:
            Strng = getattr(CF,"Fix"+Source.replace('-', '_'))(Strng,Source,reverse=True)
        except AttributeError:
            pass

        tr = gimeltra.Transliterator()

        Strng = tr.tr(Strng, sc=Source, to_sc='Latn')
        #print(Strng)
        Strng = CF.FixSemiticOutput(Strng, Source, Target) ## generic fixes
        #print(Strng)
        Strng = PrP.FixSemiticRoman(Strng, Source) ## Specific fixes
        #print(Strng)

        Strng = convertScript(Strng, 'RomanSemitic', Target)

        if Source == 'Ugar':
            Strng = Strng.replace('𐎟', ' ') # reverse ugaritic word separator

    #print(Strng)

    Strng = PP.default(Strng)

    #print(Strng)

    return Strng


# -*- coding: utf-8 -*-

# Script Mapping for Devanagari

#http://www.laomanuscripts.net/downloads/DLLM_pali_laoromanisation.pdf

#Lao aG to aM, check for buddhaM vs buddhaG...

VowelMap =  [
            '\u0EAD',
            '\u0EAD\u0EB2',
            '\u0EAD\u0EB4',
            '\u0EAD\u0EB5',
            '\u0EAD\u0EB8',
            '\u0EAD\u0EB9',
            '\u0EA3\u0EB8\u02BD',
            '\u0EA3\u0EB9\u02BD',
            '\u0EA5\u0EB8\u02BD',
            '\u0EA5\u0EB9\u02BD',
            '\u0EAD\u0EC0',
            '\u0EAD\u0EC4',
            '\u0EAD\u0EC2',
            '\u0EAD\u0EC0\u0EBB\u0EB2',
            ]

SouthVowelMap = [
                '\u0EAD\u0EC0\u0EB0',
                '\u0EAD\u0EC2\u0EB0',
                ]

ModernVowelMap = [
                 '\u0EAD\u0EC1\u0EB0',
                 '\u0EAD\u0EC0\u0EB2\u0EB0',
                 ]

SinhalaVowelMap = [
                  '\u0EAD\u0EC1'
                  ]

VowelSignMap =  [
                 '\u0EB2',
                 '\u0EB4',
                 '\u0EB5',
                 '\u0EB8',
                 '\u0EB9',
                 '\u0EBA\u0EA3\u0EB8\u02BD',
                 '\u0EBA\u0EA3\u0EB9\u02BD',
                 '\u0EBA\u0EA5\u0EB8\u02BD',
                 '\u0EBA\u0EA5\u0EB9\u02BD',
                 '\u0EC0',
                 '\u0EC4',
                 '\u0EC2',
                 '\u0EC0\u0EBB\u0EB2',
                ]

SouthVowelSignMap = [
                     '\u0EC0\u0EB0',
                     '\u0EC2\u0EB0',
                    ]

ModernVowelSignMap =[
                     '\u0EC1\u0EB0',
                     '\u0EC0\u0EB2\u0EB0',
                    ]

SinhalaVowelSignMap = [
                       '\u0EC1'
                      ]

AyogavahaMap = [
               '\u0ECD\u02BD',
               '\u0ECD',
               '\u0EAB\u0EBA\u02BD'
               ]

ViramaMap =  [
             '\u0EBA'
             ]

ConsonantMap =  [
                '\u0E81',
                '\u0E82',
                '\u0E84',
                '\u0E84\u02BD',
                '\u0E87',

                '\u0E88',
                '\u0E88\u02BD',
                '\u0E8A',
                '\u0E8A\u02BD',
                '\u0E8D',

                '\u0E95\u02BD',
                '\u0E96\u02BD',
                '\u0E97\u02BD',
                '\u0E97\u02BD',
                '\u0E99\u02BD',

                '\u0E95',
                '\u0E96',
                '\u0E97',
                '\u0E97\u02BD',
                '\u0E99',

                '\u0E9B',
                '\u0E9C',
                '\u0E9E',
                '\u0E9E\u02BD',
                '\u0EA1',

                '\u0EA2',
                '\u0EA3',
                '\u0EA5',
                '\u0EA7',

                '\u0EAA\u02BD',
                '\u0EAA\u02BD',
                '\u0EAA',
                '\u0EAB'
                ]

SouthConsonantMap = [
                    '\u0EA5\u02BD',
                    '\u0EA5\u02BD',
                    'ຮ',
                    '\u0E99\u02BD'
                    ]

NuktaConsonantMap =  [
                    '\u0E81\u02BD',
                    '\u0E82\u02BD',
                    '\u0E84\u02BD',
                    '\u0E8A\u02BD',
                    '\u0E97\u02BD',
                    '\u0E97\u02BD',
                    'ຝ',
                    '\u0EA2\u02BD',
                     ]

SinhalaConsonantMap =[
                     '\u0ECD\u0E84\u02BD',
                     '\u0ECD\u0E8A\u02BD',
                     '\u0ECD\u0E97\u02BD',
                     '\u0ECD\u0E97\u02BD',
                     '\u0ECD\u0E9E\u02BD',
                      ]

NuktaMap = [
           '\uEB0A'
           ]

OmMap = [
        '\u0EC2\u0EAD\u0E87\u0EBA'
        ]

SignMap =[
         "'",
         '\u0EAF',
         '\u0EAF\u0EAF'
         ]

Aytham =[AyogavahaMap[2]+'\u02BD']

NumeralMap = [
             '\u0ED0',
             '\u0ED1',
             '\u0ED2',
             '\u0ED3',
             '\u0ED4',
             '\u0ED5',
             '\u0ED6',
             '\u0ED7',
             '\u0ED8',
             '\u0ED9'
             ]

from ... import GeneralMap as GM

import os
GM.add_additional_chars(dict([(charlist, globals()[charlist]) for charlist in GM.CharmapLists]),
                        os.path.splitext(__file__)[0].split(os.path.sep)[-1])
